@@ -10,23 +10,26 @@
  * @brief standard mathematical functions for xexpressions
  */
 
-#ifndef XBUILDER_HPP
-#define XBUILDER_HPP
+#ifndef XTENSOR_BUILDER_HPP
+#define XTENSOR_BUILDER_HPP
 
 #include <array>
-#include <cstddef>
 #include <cmath>
+#include <cstddef>
 #include <functional>
 #include <utility>
 #include <vector>
+#ifdef X_OLD_CLANG
+    #include <initializer_list>
+#endif
+
+#include "xtl/xclosure.hpp"
+#include "xtl/xsequence.hpp"
 
 #include "xbroadcast.hpp"
 #include "xfunction.hpp"
 #include "xgenerator.hpp"
-
-#ifdef X_OLD_CLANG
-    #include <initializer_list>
-#endif
+#include "xoperation.hpp"
 
 namespace xt
 {
@@ -188,7 +191,8 @@ namespace xt
             template <class It>
             inline T operator()(const It& /*begin*/, const It& end) const
             {
-                return *(end - 1) == *(end - 2) + m_k ? T(1) : T(0);
+                using lvalue_type = typename std::iterator_traits<It>::value_type;
+                return *(end - 1) == *(end - 2) + static_cast<lvalue_type>(static_cast<unsigned int>(m_k)) ? T(1) : T(0);
             }
 
         private:
@@ -267,8 +271,8 @@ namespace xt
     template <class T>
     inline auto linspace(T start, T stop, std::size_t num_samples = 50, bool endpoint = true) noexcept
     {
-        T step = (stop - start) / T(num_samples - (endpoint ? 1 : 0));
-        return detail::make_xgenerator(detail::arange_impl<T>(start, stop, step), {num_samples});
+        double step = double(stop - start) / double(num_samples - (endpoint ? 1 : 0));
+        return cast<T>(detail::make_xgenerator(detail::arange_impl<double>(double(start), double(stop), step), {num_samples}));
     }
 
     /**
@@ -284,7 +288,7 @@ namespace xt
     template <class T>
     inline auto logspace(T start, T stop, std::size_t num_samples, T base = 10, bool endpoint = true) noexcept
     {
-        return pow(std::forward<T>(base), linspace(start, stop, num_samples, endpoint));
+        return cast<T>(pow(std::move(base), linspace(start, stop, num_samples, endpoint)));
     }
 
     namespace detail
@@ -295,7 +299,7 @@ namespace xt
         public:
 
             using size_type = std::size_t;
-            using value_type = std::common_type_t<typename std::decay_t<CT>::value_type...>;
+            using value_type = promote_type_t<typename std::decay_t<CT>::value_type...>;
 
             inline concatenate_impl(std::tuple<CT...>&& t, size_type axis)
                 : m_t(t), m_axis(axis)
@@ -356,7 +360,7 @@ namespace xt
         public:
 
             using size_type = std::size_t;
-            using value_type = std::common_type_t<typename std::decay_t<CT>::value_type...>;
+            using value_type = promote_type_t<typename std::decay_t<CT>::value_type...>;
 
             inline stack_impl(std::tuple<CT...>&& t, size_type axis)
                 : m_t(t), m_axis(axis)
@@ -386,7 +390,7 @@ namespace xt
                     return arr[idx];
                 };
                 size_type i = idx[m_axis];
-                idx.erase(idx.begin() + m_axis);
+                idx.erase(idx.begin() + std::ptrdiff_t(m_axis));
                 return apply<value_type>(i, get_item, m_t);
             }
 
@@ -412,7 +416,7 @@ namespace xt
             template <class... Args>
             value_type operator()(Args... args) const
             {
-                std::array<size_type, sizeof...(Args)> args_arr({static_cast<size_type>(args)...});
+                std::array<size_type, sizeof...(Args)> args_arr = {static_cast<size_type>(args)...};
                 return m_source(args_arr[m_axis]);
             }
 
@@ -436,7 +440,7 @@ namespace xt
     template <class... Types>
     inline auto xtuple(Types&&... args)
     {
-        return std::tuple<const_closure_t<Types>...>(std::forward<Types>(args)...);
+        return std::tuple<xtl::const_closure_type_t<Types>...>(std::forward<Types>(args)...);
     }
 
     /**
@@ -458,7 +462,7 @@ namespace xt
     inline auto concatenate(std::tuple<CT...>&& t, std::size_t axis = 0)
     {
         using shape_type = promote_shape_t<typename std::decay_t<CT>::shape_type...>;
-        shape_type new_shape = forward_sequence<shape_type>(std::get<0>(t).shape());
+        shape_type new_shape = xtl::forward_sequence<shape_type>(std::get<0>(t).shape());
         auto shape_at_axis = [&axis](std::size_t prev, auto& arr) -> std::size_t {
             return prev + arr.shape()[axis];
         };
@@ -482,7 +486,7 @@ namespace xt
         inline T add_axis(T arr, std::size_t axis, std::size_t value)
         {
             T temp(arr);
-            temp.insert(temp.begin() + axis, value);
+            temp.insert(temp.begin() + std::ptrdiff_t(axis), value);
             return temp;
         }
     }
@@ -509,7 +513,7 @@ namespace xt
     inline auto stack(std::tuple<CT...>&& t, std::size_t axis = 0)
     {
         using shape_type = promote_shape_t<typename std::decay_t<CT>::shape_type...>;
-        auto new_shape = detail::add_axis(forward_sequence<shape_type>(std::get<0>(t).shape()), axis, sizeof...(CT));
+        auto new_shape = detail::add_axis(xtl::forward_sequence<shape_type>(std::get<0>(t).shape()), axis, sizeof...(CT));
         return detail::make_xgenerator(detail::stack_impl<CT...>(std::forward<std::tuple<CT...>>(t), axis), new_shape);
     }
 
@@ -580,14 +584,16 @@ namespace xt
                         idx[i] = *begin++;
                     }
                 }
+                using it_vtype = typename std::iterator_traits<It>::value_type;
+                it_vtype uoffset = static_cast<it_vtype>(m_offset);
                 if (m_offset >= 0)
                 {
                     idx[m_axis_1] = *(begin);
-                    idx[m_axis_2] = *(begin) + m_offset;
+                    idx[m_axis_2] = *(begin) + uoffset;
                 }
                 else
                 {
-                    idx[m_axis_1] = *(begin) - m_offset;
+                    idx[m_axis_1] = *(begin) - uoffset;
                     idx[m_axis_2] = *(begin);
                 }
                 return m_source[idx];
@@ -618,13 +624,15 @@ namespace xt
             template <class It>
             inline value_type operator()(It begin, It) const
             {
+                using it_vtype = typename std::iterator_traits<It>::value_type;
+                it_vtype umk = static_cast<it_vtype>(m_k);
                 if (m_k > 0)
                 {
-                    return *begin + m_k == *(begin + 1) ? m_source(*begin) : value_type(0);
+                    return *begin + umk == *(begin + 1) ? m_source(*begin) : value_type(0);
                 }
                 else
                 {
-                    return *begin + m_k == *(begin + 1) ? m_source(*begin + m_k) : value_type(0);
+                    return *begin + umk == *(begin + 1) ? m_source(*begin + umk) : value_type(0);
                 }
             }
 
@@ -669,7 +677,8 @@ namespace xt
             template <class It>
             inline value_type access_impl(It begin, It end) const
             {
-                auto it = begin + m_axis;
+                using difference_type = typename std::iterator_traits<It>::difference_type;
+                auto it = begin + difference_type(m_axis);
                 *it = m_shape_at_axis - *it;
                 return m_source.element(begin, end);
             }
@@ -727,19 +736,19 @@ namespace xt
 
     /**
      * @brief Returns the elements on the diagonal of arr
-     * If arr has more than two dimensions, then the axes specified by 
-     * axis_1 and axis_2 are used to determine the 2-D sub-array whose 
-     * diagonal is returned. The shape of the resulting array can be 
-     * determined by removing axis1 and axis2 and appending an index 
+     * If arr has more than two dimensions, then the axes specified by
+     * axis_1 and axis_2 are used to determine the 2-D sub-array whose
+     * diagonal is returned. The shape of the resulting array can be
+     * determined by removing axis1 and axis2 and appending an index
      * to the right equal to the size of the resulting diagonals.
      *
      * @param arr the input array
      * @param offset offset of the diagonal from the main diagonal. Can
      *               be positive or negative.
-     * @param axis_1 Axis to be used as the first axis of the 2-D sub-arrays 
-     *               from which the diagonals should be taken. 
-     * @param axis_2 Axis to be used as the second axis of the 2-D sub-arrays 
-     *               from which the diagonals should be taken. 
+     * @param axis_1 Axis to be used as the first axis of the 2-D sub-arrays
+     *               from which the diagonals should be taken.
+     * @param axis_2 Axis to be used as the second axis of the 2-D sub-arrays
+     *               from which the diagonals should be taken.
      * @returns xexpression with values of the diagonal
      *
      * \code{.cpp}
@@ -760,13 +769,13 @@ namespace xt
 
         // The following shape calculation code is an almost verbatim adaptation of numpy:
         // https://github.com/numpy/numpy/blob/2aabeafb97bea4e1bfa29d946fbf31e1104e7ae0/numpy/core/src/multiarray/item_selection.c#L1799
-        auto ret_shape = make_sequence<shape_type>(dimension - 1, 0);
-        std::size_t dim_1 = shape[axis_1];
-        std::size_t dim_2 = shape[axis_2];
+        auto ret_shape = xtl::make_sequence<shape_type>(dimension - 1, 0);
+        int dim_1 = static_cast<int>(shape[axis_1]);
+        int dim_2 = static_cast<int>(shape[axis_2]);
 
         offset >= 0 ? dim_2 -= offset : dim_1 += offset;
 
-        auto diag_size = dim_2 < dim_1 ? dim_2 : dim_1;
+        auto diag_size = std::size_t(dim_2 < dim_1 ? dim_2 : dim_1);
 
         std::size_t i = 0;
         for (std::size_t idim = 0; idim < dimension; ++idim)
@@ -801,7 +810,8 @@ namespace xt
     inline auto diag(E&& arr, int k = 0)
     {
         using CT = xclosure_t<E>;
-        std::size_t s = arr.shape()[0] + std::abs(k);
+        std::size_t sk = std::size_t(std::abs(k));
+        std::size_t s = arr.shape()[0] + sk;
         return detail::make_xgenerator(detail::fn_impl<detail::diag_fn<CT>>(detail::diag_fn<CT>(std::forward<E>(arr), k)),
                                        {s, s});
     }
@@ -810,7 +820,7 @@ namespace xt
      * @brief Reverse the order of elements in an xexpression along the given axis.
      * Note: A NumPy/Matlab style `flipud(arr)` is equivalent to `xt::flip(arr, 0)`,
      * `fliplr(arr)` to `xt::flip(arr, 1)`.
-     * 
+     *
      * @param arr the input xexpression
      * @param axis the axis along which elements should be reversed
      *
