@@ -6,13 +6,16 @@
 * The full license is in the file LICENSE, distributed with this software. *
 ****************************************************************************/
 
-#ifndef XCONTAINER_HPP
-#define XCONTAINER_HPP
+#ifndef XTENSOR_CONTAINER_HPP
+#define XTENSOR_CONTAINER_HPP
 
 #include <algorithm>
 #include <functional>
 #include <numeric>
 #include <stdexcept>
+#include <iostream>
+
+#include "xtl/xsequence.hpp"
 
 #include "xiterable.hpp"
 #include "xiterator.hpp"
@@ -28,14 +31,12 @@ namespace xt
     {
         using inner_shape_type = typename xcontainer_inner_types<D>::inner_shape_type;
         using container_type = typename xcontainer_inner_types<D>::container_type;
-        using iterator = typename container_type::iterator;
-        using const_iterator = typename container_type::const_iterator;
-        using reverse_iterator = typename container_type::reverse_iterator;
-        using const_reverse_iterator = typename container_type::const_reverse_iterator;
         using stepper = xstepper<D>;
         using const_stepper = xstepper<const D>;
     };
 
+#define DL DEFAULT_LAYOUT
+
     /**
      * @class xcontainer
      * @brief Base class for dense multidimensional containers.
@@ -48,7 +49,7 @@ namespace xt
      *           provides the interface.
      */
     template <class D>
-    class xcontainer : public xiterable<D>
+    class xcontainer : private xiterable<D>
     {
     public:
 
@@ -63,6 +64,7 @@ namespace xt
         using const_pointer = typename container_type::const_pointer;
         using size_type = typename container_type::size_type;
         using difference_type = typename container_type::difference_type;
+        using simd_value_type = xsimd::simd_type<value_type>;
 
         using shape_type = typename inner_types::shape_type;
         using strides_type = typename inner_types::strides_type;
@@ -73,15 +75,13 @@ namespace xt
         using inner_backstrides_type = typename inner_types::inner_backstrides_type;
 
         using iterable_base = xiterable<D>;
-
-        using iterator = typename iterable_base::iterator;
-        using const_iterator = typename iterable_base::const_iterator;
-
         using stepper = typename iterable_base::stepper;
         using const_stepper = typename iterable_base::const_stepper;
 
-        using reverse_iterator = typename iterable_base::reverse_iterator;
-        using const_reverse_iterator = typename iterable_base::const_reverse_iterator;
+        static constexpr layout_type static_layout = inner_types::layout;
+        static constexpr bool contiguous_layout = static_layout != layout_type::dynamic;
+        using data_alignment = xsimd::container_alignment<container_type>;
+        using simd_type = xsimd::simd_type<value_type>;
 
         size_type size() const noexcept;
 
@@ -97,6 +97,12 @@ namespace xt
         template <class... Args>
         const_reference operator()(Args... args) const;
 
+        template <class... Args>
+        reference at(Args... args);
+
+        template <class... Args>
+        const_reference at(Args... args) const;
+
         reference operator[](const xindex& index);
         reference operator[](size_type i);
         const_reference operator[](const xindex& index) const;
@@ -120,22 +126,6 @@ namespace xt
         template <class S>
         bool is_trivial_broadcast(const S& strides) const noexcept;
 
-        iterator begin() noexcept;
-        iterator end() noexcept;
-
-        const_iterator begin() const noexcept;
-        const_iterator end() const noexcept;
-        const_iterator cbegin() const noexcept;
-        const_iterator cend() const noexcept;
-
-        reverse_iterator rbegin() noexcept;
-        reverse_iterator rend() noexcept;
-
-        const_reverse_iterator rbegin() const noexcept;
-        const_reverse_iterator rend() const noexcept;
-        const_reverse_iterator crbegin() const noexcept;
-        const_reverse_iterator crend() const noexcept;
-
         template <class S>
         stepper stepper_begin(const S& shape) noexcept;
         template <class S>
@@ -146,12 +136,154 @@ namespace xt
         template <class S>
         const_stepper stepper_end(const S& shape, layout_type l) const noexcept;
 
-        using container_iterator = typename container_type::iterator;
-        using const_container_iterator = typename container_type::const_iterator;
-
         reference data_element(size_type i);
         const_reference data_element(size_type i) const;
 
+        template <class align, class simd = simd_value_type>
+        void store_simd(size_type i, const simd& e);
+        template <class align, class simd = simd_value_type>
+        simd load_simd(size_type i) const;
+
+#if defined(_MSC_VER) && _MSC_VER >= 1910
+        // Workaround for compiler bug in Visual Studio 2017 with respect to alias templates with non-type parameters.
+        template <layout_type L>
+        using layout_iterator = xiterator<typename iterable_base::stepper, typename iterable_base::inner_shape_type*, L>;
+        template <layout_type L>
+        using const_layout_iterator = xiterator<typename iterable_base::const_stepper, typename iterable_base::inner_shape_type*, L>;
+        template <layout_type L>
+        using reverse_layout_iterator = std::reverse_iterator<layout_iterator<L>>;
+        template <layout_type L>
+        using const_reverse_layout_iterator = std::reverse_iterator<const_layout_iterator<L>>;
+#else
+        template <layout_type L>
+        using layout_iterator = typename iterable_base::template layout_iterator<L>;
+        template <layout_type L>
+        using const_layout_iterator = typename iterable_base::template const_layout_iterator<L>;
+        template <layout_type L>
+        using reverse_layout_iterator = typename iterable_base::template reverse_layout_iterator<L>;
+        template <layout_type L>
+        using const_reverse_layout_iterator = typename iterable_base::template const_reverse_layout_iterator<L>;
+#endif
+
+        template <class S, layout_type L>
+        using broadcast_iterator = typename iterable_base::template broadcast_iterator<S, L>;
+        template <class S, layout_type L>
+        using const_broadcast_iterator = typename iterable_base::template const_broadcast_iterator<S, L>;
+        template <class S, layout_type L>
+        using reverse_broadcast_iterator = typename iterable_base::template reverse_broadcast_iterator<S, L>;
+        template <class S, layout_type L>
+        using const_reverse_broadcast_iterator = typename iterable_base::template const_reverse_broadcast_iterator<S, L>;
+
+        using storage_iterator = typename container_type::iterator;
+        using const_storage_iterator = typename container_type::const_iterator;
+        using reverse_storage_iterator = typename container_type::reverse_iterator;
+        using const_reverse_storage_iterator = typename container_type::const_reverse_iterator;
+
+        template <layout_type L, class It1, class It2>
+        using select_iterator_impl = std::conditional_t<L == static_layout, It1, It2>;
+
+        template <layout_type L>
+        using select_iterator = select_iterator_impl<L, storage_iterator, layout_iterator<L>>;
+        template <layout_type L>
+        using select_const_iterator = select_iterator_impl<L, const_storage_iterator, const_layout_iterator<L>>;
+        template <layout_type L>
+        using select_reverse_iterator = select_iterator_impl<L, reverse_storage_iterator, reverse_layout_iterator<L>>;
+        template <layout_type L>
+        using select_const_reverse_iterator = select_iterator_impl<L, const_reverse_storage_iterator, const_reverse_layout_iterator<L>>;
+
+        using iterator = select_iterator<DL>;
+        using const_iterator = select_const_iterator<DL>;
+        using reverse_iterator = select_reverse_iterator<DL>;
+        using const_reverse_iterator = select_const_reverse_iterator<DL>;
+
+        template <layout_type L = DL>
+        select_iterator<L> begin() noexcept;
+        template <layout_type L = DL>
+        select_iterator<L> end() noexcept;
+
+        template <layout_type L = DL>
+        select_const_iterator<L> begin() const noexcept;
+        template <layout_type L = DL>
+        select_const_iterator<L> end() const noexcept;
+        template <layout_type L = DL>
+        select_const_iterator<L> cbegin() const noexcept;
+        template <layout_type L = DL>
+        select_const_iterator<L> cend() const noexcept;
+
+        template <layout_type L = DL>
+        select_reverse_iterator<L> rbegin() noexcept;
+        template <layout_type L = DL>
+        select_reverse_iterator<L> rend() noexcept;
+
+        template <layout_type L = DL>
+        select_const_reverse_iterator<L> rbegin() const noexcept;
+        template <layout_type L = DL>
+        select_const_reverse_iterator<L> rend() const noexcept;
+        template <layout_type L = DL>
+        select_const_reverse_iterator<L> crbegin() const noexcept;
+        template <layout_type L = DL>
+        select_const_reverse_iterator<L> crend() const noexcept;
+
+        template <class S, layout_type L = DL>
+        broadcast_iterator<S, L> begin(const S& shape) noexcept;
+        template <class S, layout_type L = DL>
+        broadcast_iterator<S, L> end(const S& shape) noexcept;
+
+        template <class S, layout_type L = DL>
+        const_broadcast_iterator<S, L> begin(const S& shape) const noexcept;
+        template <class S, layout_type L = DL>
+        const_broadcast_iterator<S, L> end(const S& shape) const noexcept;
+        template <class S, layout_type L = DL>
+        const_broadcast_iterator<S, L> cbegin(const S& shape) const noexcept;
+        template <class S, layout_type L = DL>
+        const_broadcast_iterator<S, L> cend(const S& shape) const noexcept;
+
+
+        template <class S, layout_type L = DL>
+        reverse_broadcast_iterator<S, L> rbegin(const S& shape) noexcept;
+        template <class S, layout_type L = DL>
+        reverse_broadcast_iterator<S, L> rend(const S& shape) noexcept;
+
+        template <class S, layout_type L = DL>
+        const_reverse_broadcast_iterator<S, L> rbegin(const S& shape) const noexcept;
+        template <class S, layout_type L = DL>
+        const_reverse_broadcast_iterator<S, L> rend(const S& shape) const noexcept;
+        template <class S, layout_type L = DL>
+        const_reverse_broadcast_iterator<S, L> crbegin(const S& shape) const noexcept;
+        template <class S, layout_type L = DL>
+        const_reverse_broadcast_iterator<S, L> crend(const S& shape) const noexcept;
+
+        template <layout_type L = DL>
+        storage_iterator storage_begin() noexcept;
+        template <layout_type L = DL>
+        storage_iterator storage_end() noexcept;
+
+        template <layout_type L = DL>
+        const_storage_iterator storage_begin() const noexcept;
+        template <layout_type L = DL>
+        const_storage_iterator storage_end() const noexcept;
+        template <layout_type L = DL>
+        const_storage_iterator storage_cbegin() const noexcept;
+        template <layout_type L = DL>
+        const_storage_iterator storage_cend() const noexcept;
+
+        template <layout_type L = DL>
+        reverse_storage_iterator storage_rbegin() noexcept;
+        template <layout_type L = DL>
+        reverse_storage_iterator storage_rend() noexcept;
+
+        template <layout_type L = DL>
+        const_reverse_storage_iterator storage_rbegin() const noexcept;
+        template <layout_type L = DL>
+        const_reverse_storage_iterator storage_rend() const noexcept;
+        template <layout_type L = DL>
+        const_reverse_storage_iterator storage_crbegin() const noexcept;
+        template <layout_type L = DL>
+        const_reverse_storage_iterator storage_crend() const noexcept;
+
+        using container_iterator = storage_iterator;
+        using const_container_iterator = const_storage_iterator;
+
     protected:
 
         xcontainer() = default;
@@ -170,6 +302,9 @@ namespace xt
 
     private:
 
+        friend class xiterable<D>;
+        friend class xconst_iterable<D>;
+
         template <class C>
         friend class xstepper;
 
@@ -184,6 +319,8 @@ namespace xt
         const derived_type& derived_cast() const;
     };
 
+#undef DL
+
     /**
      * @class xstrided_container
      * @brief Partial implementation of xcontainer that embeds the strides and the shape
@@ -195,9 +332,8 @@ namespace xt
      *
      * @tparam D The derived type, i.e. the inheriting class for which xstrided_container
      *           provides the partial imlpementation of xcontainer.
-     * @tparam L The layout_type of the xstrided_container.
      */
-    template <class D, layout_type L>
+    template <class D>
     class xstrided_container : public xcontainer<D>
     {
     public:
@@ -216,15 +352,12 @@ namespace xt
         using inner_strides_type = typename base_type::inner_strides_type;
         using inner_backstrides_type = typename base_type::inner_backstrides_type;
 
-        static constexpr layout_type static_layout = L;
-        static constexpr bool contiguous_layout = static_layout != layout_type::dynamic;
-
         template <class S = shape_type>
-        void reshape(const S& shape, bool force = false);
+        void reshape(S&& shape, bool force = false);
         template <class S = shape_type>
-        void reshape(const S& shape, layout_type l);
+        void reshape(S&& shape, layout_type l);
         template <class S = shape_type>
-        void reshape(const S& shape, const strides_type& strides);
+        void reshape(S&& shape, const strides_type& strides);
 
         layout_type layout() const noexcept;
 
@@ -255,7 +388,7 @@ namespace xt
         inner_shape_type m_shape;
         inner_strides_type m_strides;
         inner_backstrides_type m_backstrides;
-        layout_type m_layout = L;
+        layout_type m_layout = base_type::static_layout;
     };
 
     /******************************
@@ -266,16 +399,7 @@ namespace xt
     template <class It>
     inline It xcontainer<D>::data_xend_impl(It end, layout_type l) const noexcept
     {
-        if (dimension() == 0)
-        {
-            return end;
-        }
-        else
-        {
-            auto leading_stride = (l == layout_type::row_major ? strides().back() : strides().front());
-            leading_stride = std::max(leading_stride, typename inner_strides_type::value_type(1));
-            return end - 1 + leading_stride;
-        }
+        return strided_data_end(*this, end, l);
     }
 
     template <class D>
@@ -318,7 +442,7 @@ namespace xt
     template <class D>
     inline auto xcontainer<D>::size() const noexcept -> size_type
     {
-        return data().size();
+        return compute_size(shape());
     }
 
     /**
@@ -358,7 +482,6 @@ namespace xt
     }
     //@}
 
-
     /**
      * @name Data
      */
@@ -374,6 +497,7 @@ namespace xt
     inline auto xcontainer<D>::operator()(Args... args) -> reference
     {
         XTENSOR_ASSERT(check_index(shape(), args...));
+        XTENSOR_CHECK_DIMENSION(shape(), args...);
         size_type index = data_offset<size_type>(strides(), static_cast<size_type>(args)...);
         return data()[index];
     }
@@ -389,10 +513,45 @@ namespace xt
     inline auto xcontainer<D>::operator()(Args... args) const -> const_reference
     {
         XTENSOR_ASSERT(check_index(shape(), args...));
+        XTENSOR_CHECK_DIMENSION(shape(), args...);
         size_type index = data_offset<size_type>(strides(), static_cast<size_type>(args)...);
         return data()[index];
     }
 
+    /**
+     * Returns a reference to the element at the specified position in the container,
+     * after dimension and bounds checking.
+     * @param args a list of indices specifying the position in the container. Indices
+     * must be unsigned integers, the number of indices should be equal to the number of dimensions
+     * of the container.
+     * @exception std::out_of_range if the number of argument is greater than the number of dimensions
+     * or if indices are out of bounds.
+     */
+    template <class D>
+    template <class... Args>
+    inline auto xcontainer<D>::at(Args... args) -> reference
+    {
+        check_access(shape(), static_cast<size_type>(args)...);
+        return this->operator()(args...);
+    }
+
+    /**
+     * Returns a constant reference to the element at the specified position in the container,
+     * after dimension and bounds checking.
+     * @param args a list of indices specifying the position in the container. Indices
+     * must be unsigned integers, the number of indices should be equal to the number of dimensions
+     * of the container.
+     * @exception std::out_of_range if the number of argument is greater than the number of dimensions
+     * or if indices are out of bounds.
+     */
+    template <class D>
+    template <class... Args>
+    inline auto xcontainer<D>::at(Args... args) const -> const_reference
+    {
+        check_access(shape(), static_cast<size_type>(args)...);
+        return this->operator()(args...);
+    }
+
     /**
      * Returns a reference to the element at the specified position in the container.
      * @param index a sequence of indices specifying the position in the container. Indices
@@ -534,138 +693,316 @@ namespace xt
     //@}
 
     /****************
-     * iterator api *
+     * Iterator api *
      ****************/
 
-    /**
-     * @name Iterators
-     */
-    //@{
-    /**
-     * Returns an iterator to the first element of the buffer containing
-     * the elements of the container.
-     */
     template <class D>
-    inline auto xcontainer<D>::begin() noexcept -> iterator
+    template <layout_type L>
+    inline auto xcontainer<D>::begin() noexcept -> select_iterator<L>
+    {
+        return static_if<L == static_layout>([&](auto self)
+        {
+            return self(*this).template storage_begin<L>();
+        }, /*else*/ [&](auto self)
+        {
+            return self(*this).iterable_base::template begin<L>();
+        });
+    }
+
+    template <class D>
+    template <layout_type L>
+    inline auto xcontainer<D>::end() noexcept -> select_iterator<L>
+    {
+        return static_if<L == static_layout>([&](auto self)
+        {
+            return self(*this).template storage_end<L>();
+        }, /*else*/ [&](auto self)
+        {
+            return self(*this).iterable_base::template end<L>();
+        });
+    }
+
+    template <class D>
+    template <layout_type L>
+    inline auto xcontainer<D>::begin() const noexcept -> select_const_iterator<L>
+    {
+        return cbegin<L>();
+    }
+
+    template <class D>
+    template <layout_type L>
+    inline auto xcontainer<D>::end() const noexcept -> select_const_iterator<L>
+    {
+        return cend<L>();
+    }
+
+    template <class D>
+    template <layout_type L>
+    inline auto xcontainer<D>::cbegin() const noexcept -> select_const_iterator<L>
+    {
+        return static_if<L == static_layout>([&](auto self)
+        {
+            return self(*this).template storage_cbegin<L>();
+        }, /*else*/ [&](auto self)
+        {
+            return self(*this).iterable_base::template cbegin<L>();
+        });
+    }
+
+    template <class D>
+    template <layout_type L>
+    inline auto xcontainer<D>::cend() const noexcept -> select_const_iterator<L>
+    {
+        return static_if<L == static_layout>([&](auto self)
+        {
+            return self(*this).template storage_cend<L>();
+        }, /*else*/ [&](auto self)
+        {
+            return self(*this).iterable_base::template cend<L>();
+        });
+    }
+
+    template <class D>
+    template <layout_type L>
+    inline auto xcontainer<D>::rbegin() noexcept -> select_reverse_iterator<L>
+    {
+        return static_if<L == static_layout>([&](auto self)
+        {
+            return self(*this).template storage_rbegin<L>();
+        }, /*else*/ [&](auto self)
+        {
+            return self(*this).iterable_base::template rbegin<L>();
+        });
+    }
+
+    template <class D>
+    template <layout_type L>
+    inline auto xcontainer<D>::rend() noexcept -> select_reverse_iterator<L>
+    {
+        return static_if<L == static_layout>([&](auto self)
+        {
+            return self(*this).template storage_rend<L>();
+        }, /*else*/ [&](auto self)
+        {
+            return self(*this).iterable_base::template rend<L>();
+        });
+    }
+
+    template <class D>
+    template <layout_type L>
+    inline auto xcontainer<D>::rbegin() const noexcept -> select_const_reverse_iterator<L>
+    {
+        return crbegin<L>();
+    }
+
+    template <class D>
+    template <layout_type L>
+    inline auto xcontainer<D>::rend() const noexcept -> select_const_reverse_iterator<L>
+    {
+        return crend<L>();
+    }
+
+    template <class D>
+    template <layout_type L>
+    inline auto xcontainer<D>::crbegin() const noexcept -> select_const_reverse_iterator<L>
+    {
+        return static_if<L == static_layout>([&](auto self)
+        {
+            return self(*this).template storage_crbegin<L>();
+        }, /*else*/ [&](auto self)
+        {
+            return self(*this).iterable_base::template crbegin<L>();
+        });
+    }
+
+    template <class D>
+    template <layout_type L>
+    inline auto xcontainer<D>::crend() const noexcept -> select_const_reverse_iterator<L>
+    {
+        return static_if<L == static_layout>([&](auto self)
+        {
+            return self(*this).template storage_crend<L>();
+        }, /*else*/ [&](auto self)
+        {
+            return self(*this).iterable_base::template crend<L>();
+        });
+    }
+
+    /*****************************
+     * Broadcasting iterator api *
+     *****************************/
+
+    template <class D>
+    template <class S, layout_type L>
+    inline auto xcontainer<D>::begin(const S& shape) noexcept -> broadcast_iterator<S, L>
+    {
+        return iterable_base::template begin<S, L>(shape);
+    }
+
+    template <class D>
+    template <class S, layout_type L>
+    inline auto xcontainer<D>::end(const S& shape) noexcept -> broadcast_iterator<S, L>
+    {
+        return iterable_base::template end<S, L>(shape);
+    }
+
+    template <class D>
+    template <class S, layout_type L>
+    inline auto xcontainer<D>::begin(const S& shape) const noexcept -> const_broadcast_iterator<S, L>
+    {
+        return iterable_base::template begin<S, L>(shape);
+    }
+
+    template <class D>
+    template <class S, layout_type L>
+    inline auto xcontainer<D>::end(const S& shape) const noexcept -> const_broadcast_iterator<S, L>
+    {
+        return iterable_base::template end<S, L>(shape);
+    }
+
+    template <class D>
+    template <class S, layout_type L>
+    inline auto xcontainer<D>::cbegin(const S& shape) const noexcept -> const_broadcast_iterator<S, L>
+    {
+        return iterable_base::template cbegin<S, L>(shape);
+    }
+
+    template <class D>
+    template <class S, layout_type L>
+    inline auto xcontainer<D>::cend(const S& shape) const noexcept -> const_broadcast_iterator<S, L>
+    {
+        return iterable_base::template cend<S, L>(shape);
+    }
+
+    template <class D>
+    template <class S, layout_type L>
+    inline auto xcontainer<D>::rbegin(const S& shape) noexcept -> reverse_broadcast_iterator<S, L>
+    {
+        return iterable_base::template rbegin<S, L>(shape);
+    }
+
+    template <class D>
+    template <class S, layout_type L>
+    inline auto xcontainer<D>::rend(const S& shape) noexcept -> reverse_broadcast_iterator<S, L>
+    {
+        return iterable_base::template rend<S, L>(shape);
+    }
+
+    template <class D>
+    template <class S, layout_type L>
+    inline auto xcontainer<D>::rbegin(const S& shape) const noexcept -> const_reverse_broadcast_iterator<S, L>
+    {
+        return iterable_base::template rbegin<S, L>(shape);
+    }
+
+    template <class D>
+    template <class S, layout_type L>
+    inline auto xcontainer<D>::rend(const S& shape) const noexcept -> const_reverse_broadcast_iterator<S, L>
+    {
+        return iterable_base::template rend<S, L>(shape);
+    }
+
+    template <class D>
+    template <class S, layout_type L>
+    inline auto xcontainer<D>::crbegin(const S& shape) const noexcept -> const_reverse_broadcast_iterator<S, L>
+    {
+        return iterable_base::template crbegin<S, L>(shape);
+    }
+
+    template <class D>
+    template <class S, layout_type L>
+    inline auto xcontainer<D>::crend(const S& shape) const noexcept -> const_reverse_broadcast_iterator<S, L>
+    {
+        return iterable_base::template crend<S, L>(shape);
+    }
+
+    /***********************
+     * Linear iterator api *
+     ***********************/
+
+    template <class D>
+    template <layout_type L>
+    inline auto xcontainer<D>::storage_begin() noexcept -> storage_iterator
     {
         return data().begin();
     }
 
-    /**
-     * Returns an iterator to the element following the last element of
-     * the buffer containing the elements of the container.
-     */
     template <class D>
-    inline auto xcontainer<D>::end() noexcept -> iterator
+    template <layout_type L>
+    inline auto xcontainer<D>::storage_end() noexcept -> storage_iterator
     {
         return data().end();
     }
 
-    /**
-     * Returns a constant iterator to the first element of the buffer
-     * containing the elements of the container.
-     */
     template <class D>
-    inline auto xcontainer<D>::begin() const noexcept -> const_iterator
+    template <layout_type L>
+    inline auto xcontainer<D>::storage_begin() const noexcept -> const_storage_iterator
     {
-        return cbegin();
+        return data().begin();
     }
 
-    /**
-     * Returns a constant iterator to the element following the last
-     * element of the buffer containing the elements of the container.
-     */
     template <class D>
-    inline auto xcontainer<D>::end() const noexcept -> const_iterator
+    template <layout_type L>
+    inline auto xcontainer<D>::storage_end() const noexcept -> const_storage_iterator
     {
-        return cend();
+        return data().end();
     }
 
-    /**
-     * Returns a constant iterator to the first element of the buffer
-     * containing the elements of the container.
-     */
     template <class D>
-    inline auto xcontainer<D>::cbegin() const noexcept -> const_iterator
+    template <layout_type L>
+    inline auto xcontainer<D>::storage_cbegin() const noexcept -> const_storage_iterator
     {
         return data().cbegin();
     }
 
-    /**
-     * Returns a constant iterator to the element following the last
-     * element of the buffer containing the elements of the container.
-     */
     template <class D>
-    inline auto xcontainer<D>::cend() const noexcept -> const_iterator
+    template <layout_type L>
+    inline auto xcontainer<D>::storage_cend() const noexcept -> const_storage_iterator
     {
         return data().cend();
     }
-    //@}
 
-    /**
-     * @name Reverse iterators
-     */
-    //@{
-    /**
-     * Returns an iterator to the first element of the reversed buffer containing
-     * the elements of the container.
-     */
     template <class D>
-    inline auto xcontainer<D>::rbegin() noexcept -> reverse_iterator
+    template <layout_type L>
+    inline auto xcontainer<D>::storage_rbegin() noexcept -> reverse_storage_iterator
     {
         return data().rbegin();
     }
 
-    /**
-     * Returns an iterator to the element following the last element of
-     * the reversed buffer containing the elements of the container.
-     */
     template <class D>
-    inline auto xcontainer<D>::rend() noexcept -> reverse_iterator
+    template <layout_type L>
+    inline auto xcontainer<D>::storage_rend() noexcept -> reverse_storage_iterator
     {
         return data().rend();
     }
 
-    /**
-     * Returns a constant iterator to the first element of the reversed
-     * buffer containing the elements of the container.
-     */
     template <class D>
-    inline auto xcontainer<D>::rbegin() const noexcept -> const_reverse_iterator
+    template <layout_type L>
+    inline auto xcontainer<D>::storage_rbegin() const noexcept -> const_reverse_storage_iterator
     {
         return data().rbegin();
     }
 
-    /**
-     * Returns a constant iterator to the element following the last
-     * element of the reversed buffer containing the elements of the container.
-     */
     template <class D>
-    inline auto xcontainer<D>::rend() const noexcept -> const_reverse_iterator
+    template <layout_type L>
+    inline auto xcontainer<D>::storage_rend() const noexcept -> const_reverse_storage_iterator
     {
         return data().rend();
     }
 
-    /**
-     * Returns a constant iterator to the first element of the reversed
-     * buffer containing the elements of the container.
-     */
     template <class D>
-    inline auto xcontainer<D>::crbegin() const noexcept -> const_reverse_iterator
+    template <layout_type L>
+    inline auto xcontainer<D>::storage_crbegin() const noexcept -> const_reverse_storage_iterator
     {
         return data().crbegin();
     }
 
-    /**
-     * Returns a constant iterator to the element following the last
-     * element of the reversed buffer containing the elements of the container.
-     */
     template <class D>
-    inline auto xcontainer<D>::crend() const noexcept -> const_reverse_iterator
+    template <layout_type L>
+    inline auto xcontainer<D>::storage_crend() const noexcept -> const_reverse_storage_iterator
     {
         return data().crend();
     }
-    //@}
 
     /***************
      * stepper api *
@@ -739,57 +1076,73 @@ namespace xt
         return data()[i];
     }
 
+    template <class D>
+    template <class alignment, class simd>
+    inline void xcontainer<D>::store_simd(size_type i, const simd& e)
+    {
+        using align_mode = driven_align_mode_t<alignment, data_alignment>;
+        xsimd::store_simd<value_type, typename simd::value_type>(&(data()[i]), e, align_mode());
+    }
+
+    template <class D>
+    template <class alignment, class simd>
+    inline auto xcontainer<D>::load_simd(size_type i) const -> simd
+    {
+        using align_mode = driven_align_mode_t<alignment, data_alignment>;
+        return xsimd::load_simd<value_type, typename simd::value_type>(&(data()[i]), align_mode());
+    }
+
     /*************************************
      * xstrided_container implementation *
      *************************************/
 
-    template <class D, layout_type L>
-    inline xstrided_container<D, L>::xstrided_container() noexcept
+    template <class D>
+    inline xstrided_container<D>::xstrided_container() noexcept
         : base_type()
     {
-        m_shape = make_sequence<inner_shape_type>(base_type::dimension(), 1);
+        m_shape = xtl::make_sequence<inner_shape_type>(base_type::dimension(), 1);
     }
 
-    template <class D, layout_type L>
-    inline xstrided_container<D, L>::xstrided_container(inner_shape_type&& shape, inner_strides_type&& strides) noexcept
+    template <class D>
+    inline xstrided_container<D>::xstrided_container(inner_shape_type&& shape, inner_strides_type&& strides) noexcept
         : base_type(), m_shape(std::move(shape)), m_strides(std::move(strides))
     {
-        m_backstrides = make_sequence<inner_backstrides_type>(m_shape.size(), 0);
+        m_backstrides = xtl::make_sequence<inner_backstrides_type>(m_shape.size(), 0);
         adapt_strides(m_shape, m_strides, m_backstrides);
     }
 
-    template <class D, layout_type L>
-    inline auto xstrided_container<D, L>::shape_impl() noexcept -> inner_shape_type&
+    template <class D>
+    inline auto xstrided_container<D>::shape_impl() noexcept -> inner_shape_type&
     {
         return m_shape;
     }
 
-    template <class D, layout_type L>
-    inline auto xstrided_container<D, L>::shape_impl() const noexcept -> const inner_shape_type&
+    template <class D>
+    inline auto xstrided_container<D>::shape_impl() const noexcept -> const inner_shape_type&
     {
         return m_shape;
     }
 
-    template <class D, layout_type L>
-    inline auto xstrided_container<D, L>::strides_impl() noexcept -> inner_strides_type&
+    template <class D>
+    inline auto xstrided_container<D>::strides_impl() noexcept -> inner_strides_type&
     {
         return m_strides;
     }
 
-    template <class D, layout_type L>
-    inline auto xstrided_container<D, L>::strides_impl() const noexcept -> const inner_strides_type&
+    template <class D>
+    inline auto xstrided_container<D>::strides_impl() const noexcept -> const inner_strides_type&
     {
         return m_strides;
     }
 
-    template <class D, layout_type L>
-    inline auto xstrided_container<D, L>::backstrides_impl() noexcept -> inner_backstrides_type&
+    template <class D>
+    inline auto xstrided_container<D>::backstrides_impl() noexcept -> inner_backstrides_type&
     {
         return m_backstrides;
     }
 
-    template <class D, layout_type L>
-    inline auto xstrided_container<D, L>::backstrides_impl() const noexcept -> const inner_backstrides_type&
+    template <class D>
+    inline auto xstrided_container<D>::backstrides_impl() const noexcept -> const inner_backstrides_type&
     {
         return m_backstrides;
     }
@@ -798,20 +1151,34 @@ namespace xt
      * Return the layout_type of the container
      * @return layout_type of the container
      */
-    template <class D, layout_type L>
-    layout_type xstrided_container<D, L>::layout() const noexcept
+    template <class D>
+    layout_type xstrided_container<D>::layout() const noexcept
     {
         return m_layout;
     }
 
+    namespace detail
+    {
+        template <class C, class S>
+        inline void resize_data_container(C& c, S size)
+        {
+            c.resize(size);
+        }
+
+        template <class C, class S>
+        inline void resize_data_container(const C&, S)
+        {
+        }
+    }
+
     /**
      * Reshapes the container.
      * @param shape the new shape
      * @param force force reshaping, even if the shape stays the same (default: false)
      */
-    template <class D, layout_type L>
+    template <class D>
     template <class S>
-    inline void xstrided_container<D, L>::reshape(const S& shape, bool force)
+    inline void xstrided_container<D>::reshape(S&& shape, bool force)
     {
         if (m_shape.size() != shape.size() || !std::equal(std::begin(shape), std::end(shape), std::begin(m_shape)) || force)
         {
@@ -819,11 +1186,11 @@ namespace xt
             {
                 m_layout = layout_type::row_major;  // fall back to row major
             }
-            m_shape = forward_sequence<shape_type>(shape);
+            m_shape = xtl::forward_sequence<shape_type>(shape);
             resize_container(m_strides, m_shape.size());
             resize_container(m_backstrides, m_shape.size());
             size_type data_size = compute_strides(m_shape, m_layout, m_strides, m_backstrides);
-            this->data().resize(data_size);
+            detail::resize_data_container(this->data(), data_size);
         }
     }
 
@@ -832,16 +1199,16 @@ namespace xt
      * @param shape the new shape
      * @param l the new layout_type
      */
-    template <class D, layout_type L>
+    template <class D>
     template <class S>
-    inline void xstrided_container<D, L>::reshape(const S& shape, layout_type l)
+    inline void xstrided_container<D>::reshape(S&& shape, layout_type l)
     {
-        if (L != layout_type::dynamic && l != L)
+        if (base_type::static_layout != layout_type::dynamic && l != base_type::static_layout)
         {
             throw std::runtime_error("Cannot change layout_type if template parameter not layout_type::dynamic.");
         }
         m_layout = l;
-        reshape(shape, true);
+        reshape(std::forward<S>(shape), true);
     }
 
     /**
@@ -849,20 +1216,20 @@ namespace xt
      * @param shape the new shape
      * @param strides the new strides
      */
-    template <class D, layout_type L>
+    template <class D>
     template <class S>
-    inline void xstrided_container<D, L>::reshape(const S& shape, const strides_type& strides)
+    inline void xstrided_container<D>::reshape(S&& shape, const strides_type& strides)
     {
-        if (L != layout_type::dynamic)
+        if (base_type::static_layout != layout_type::dynamic)
         {
             throw std::runtime_error("Cannot reshape with custom strides when layout() is != layout_type::dynamic.");
         }
-        m_shape = forward_sequence<shape_type>(shape);
+        m_shape = xtl::forward_sequence<shape_type>(shape);
         m_strides = strides;
         resize_container(m_backstrides, m_strides.size());
         adapt_strides(m_shape, m_strides, m_backstrides);
         m_layout = layout_type::dynamic;
-        this->data().resize(compute_size(m_shape));
+        detail::resize_data_container(this->data(), compute_size(m_shape));
     }
 }
 
