@@ -1,91 +1,242 @@
 /***************************************************************************
-* Copyright (c) 2016, Johan Mabille and Sylvain Corlay                     *
+* Copyright (c) 2016, Johan Mabille, Sylvain Corlay and Wolf Vollprecht    *
 *                                                                          *
 * Distributed under the terms of the BSD 3-Clause License.                 *
 *                                                                          *
 * The full license is in the file LICENSE, distributed with this software. *
 ****************************************************************************/
 
-#ifndef XOPERATION_HPP
-#define XOPERATION_HPP
+#ifndef XTENSOR_OPERATION_HPP
+#define XTENSOR_OPERATION_HPP
 
 #include <algorithm>
 #include <functional>
 #include <type_traits>
 
+#include "xtl/xsequence.hpp"
+
 #include "xfunction.hpp"
+//#include "xoptional.hpp"
 #include "xscalar.hpp"
 #include "xstrides.hpp"
 
 namespace xt
 {
 
+    template <class F, class R, class... CT>
+    class xoptional_function;
+
+    /***********
+     * helpers *
+     ***********/
+
+    namespace detail
+    {
+        template <class T, class R>
+        struct functor_return_type
+        {
+            using type = R;
+            using simd_type = xsimd::simd_type<R>;
+        };
+
+        template <class T>
+        struct functor_return_type<T, bool>
+        {
+            using type = bool;
+            using simd_type = xsimd::simd_bool_type<T>;
+        };
+    }
+
+#define UNARY_OPERATOR_FUNCTOR_IMPL(NAME, OP, R)                                \
+    template <class T>                                                          \
+    struct NAME                                                                 \
+    {                                                                           \
+        using return_type = xt::detail::functor_return_type<T, R>;              \
+        using argument_type = T;                                                \
+        using result_type = typename return_type::type;                         \
+        using simd_value_type = xsimd::simd_type<T>;                            \
+        using simd_result_type = typename return_type::simd_type;               \
+        constexpr result_type operator()(const T& arg) const                    \
+        {                                                                       \
+            return OP arg;                                                      \
+        }                                                                       \
+        constexpr simd_result_type simd_apply(const simd_value_type& arg) const \
+        {                                                                       \
+            return OP arg;                                                      \
+        }                                                                       \
+        template <class U>                                                      \
+        struct rebind                                                           \
+        {                                                                       \
+            using type = NAME<U>;                                               \
+        };                                                                      \
+    }
+
+#define UNARY_OPERATOR_FUNCTOR(NAME, OP) UNARY_OPERATOR_FUNCTOR_IMPL(NAME, OP, T)
+#define UNARY_BOOL_OPERATOR_FUNCTOR(NAME, OP) UNARY_OPERATOR_FUNCTOR_IMPL(NAME, OP, bool)
+
+    /* In this macro, T is assumed to be the promote_type of all arguments.
+       Nonetheless, operator() is implemented as a function template,
+       because automatic conversion of the actual argument types to T may
+       cause 'possible loss of data' warnings, e.g. when T is double and an
+       argument is uint64_t.
+    */
+#define BINARY_OPERATOR_FUNCTOR_IMPL(NAME, OP, R)                                \
+    template <class T>                                                           \
+    struct NAME                                                                  \
+    {                                                                            \
+        using return_type = xt::detail::functor_return_type<T, R>;               \
+        using first_argument_type = T;                                           \
+        using second_argument_type = T;                                          \
+        using result_type = typename return_type::type;                          \
+        using simd_value_type = xsimd::simd_type<T>;                             \
+        using simd_result_type = typename return_type::simd_type;                \
+        template <class T1, class T2>                                            \
+        constexpr result_type operator()(const T1& arg1, const T2& arg2) const   \
+        {                                                                        \
+            return (arg1 OP arg2);                                               \
+        }                                                                        \
+        constexpr simd_result_type simd_apply(const simd_value_type& arg1,       \
+                                              const simd_value_type& arg2) const \
+        {                                                                        \
+            return (arg1 OP arg2);                                               \
+        }                                                                        \
+        template <class U>                                                      \
+        struct rebind                                                           \
+        {                                                                       \
+            using type = NAME<U>;                                               \
+        };                                                                      \
+    }
+
+#define BINARY_OPERATOR_FUNCTOR(NAME, OP) BINARY_OPERATOR_FUNCTOR_IMPL(NAME, OP, T)
+#define BINARY_BOOL_OPERATOR_FUNCTOR(NAME, OP) BINARY_OPERATOR_FUNCTOR_IMPL(NAME, OP, bool)
+
     namespace detail
     {
 
-        /***********
-         * helpers *
-         ***********/
-
-        template <class T>
-        struct identity
-        {
-            using result_type = T;
-
-            constexpr T operator()(const T& t) const noexcept
-            {
-                return +t;
-            }
-        };
+        UNARY_OPERATOR_FUNCTOR(identity, +);
+        UNARY_OPERATOR_FUNCTOR(negate, -);
+        BINARY_OPERATOR_FUNCTOR(plus, +);
+        BINARY_OPERATOR_FUNCTOR(minus, -);
+        BINARY_OPERATOR_FUNCTOR(multiplies, *);
+        BINARY_OPERATOR_FUNCTOR(divides, /);
+        BINARY_BOOL_OPERATOR_FUNCTOR(logical_or, ||);
+        BINARY_BOOL_OPERATOR_FUNCTOR(logical_and, &&);
+        UNARY_BOOL_OPERATOR_FUNCTOR(logical_not, !);
+        BINARY_OPERATOR_FUNCTOR(bitwise_or, | );
+        BINARY_OPERATOR_FUNCTOR(bitwise_and, &);
+        BINARY_OPERATOR_FUNCTOR(bitwise_xor, ^);
+        UNARY_OPERATOR_FUNCTOR(bitwise_not, ~);
+        BINARY_BOOL_OPERATOR_FUNCTOR(less, <);
+        BINARY_BOOL_OPERATOR_FUNCTOR(less_equal, <=);
+        BINARY_BOOL_OPERATOR_FUNCTOR(greater, >);
+        BINARY_BOOL_OPERATOR_FUNCTOR(greater_equal, >=);
+        BINARY_BOOL_OPERATOR_FUNCTOR(equal_to, ==);
+        BINARY_BOOL_OPERATOR_FUNCTOR(not_equal_to, !=);
 
         template <class T>
         struct conditional_ternary
         {
             using result_type = T;
+            using simd_value_type = xsimd::simd_type<T>;
+            using simd_bool_type = xsimd::simd_bool_type<T>;
 
-            constexpr result_type operator()(const T& t1, const T& t2, const T& t3) const noexcept
+            constexpr result_type operator()(bool t1, const T& t2, const T& t3) const noexcept
             {
                 return t1 ? t2 : t3;
             }
-        };
-
-        template <class T>
-        struct minimum
-        {
-            using result_type = T;
-
-            constexpr result_type operator()(const T& t1, const T& t2) const noexcept
+            constexpr simd_value_type simd_apply(const simd_bool_type& t1,
+                                                 const simd_value_type& t2,
+                                                 const simd_value_type& t3) const noexcept
             {
-                return (t1 < t2) ? t1 : t2;
+                return xsimd::select(t1, t2, t3);
             }
-        };
-
-        template <class T>
-        struct maximum
-        {
-            using result_type = T;
-
-            constexpr result_type operator()(const T& t1, const T& t2) const noexcept
+            template <class U>
+            struct rebind
             {
-                return (t1 > t2) ? t1 : t2;
-            }
+                using type = conditional_ternary<U>;
+            };
         };
 
+        template <class R>
+        struct cast
+        {
+            template <class T>
+            struct functor
+            {
+                using return_type = xt::detail::functor_return_type<T, R>;
+                using argument_type = T;
+                using result_type = typename return_type::type;
+                using simd_value_type = xsimd::simd_type<T>;
+                using simd_result_type = typename return_type::simd_type;
+                constexpr result_type operator()(const T& arg) const
+                {
+                    return static_cast<R>(arg);
+                }
+                constexpr simd_result_type simd_apply(const simd_value_type& arg) const
+                {
+                    return static_cast<R>(arg);
+                }
+                template <class U>
+                struct rebind
+                {
+                    using type = functor<U>;
+                };
+            };
+        };
+
+        template <class Tag, class F, class... E>
+        struct select_xfunction_expression;
+
+        template <class F, class... E>
+        struct select_xfunction_expression<xtensor_expression_tag, F, E...>
+        {
+            using type = xfunction<F, typename F::result_type, E...>;
+        };
+
+        template <class F, class... E>
+        struct select_xfunction_expression<xoptional_expression_tag, F, E...>
+        {
+            using type = xoptional_function<F, typename F::result_type, E...>;
+        };
+
+        template <class Tag, class F, class... E>
+        using select_xfunction_expression_t = typename select_xfunction_expression<Tag, F, E...>::type;
+
+        template <class Tag, template <class...> class F, class... E>
+        struct build_functor_type;
+
+        template <template <class...> class F, class... E>
+        struct build_functor_type<xtensor_expression_tag, F, E...>
+        {
+            using type = F<common_value_type_t<std::decay_t<E>...>>;
+        };
+
+        template <template <class...> class F, class... E>
+        struct build_functor_type<xoptional_expression_tag, F, E...>
+        {
+            using type = F<common_value_type_t<std::decay_t<E>...>>;
+        };
+
+        template <class Tag, template <class...> class F, class... E>
+        using build_functor_type_t = typename build_functor_type<Tag, F, E...>::type;
+
         template <template <class...> class F, class... E>
         inline auto make_xfunction(E&&... e) noexcept
         {
-            using functor_type = F<common_value_type_t<std::decay_t<E>...>>;
-            using result_type = typename functor_type::result_type;
-            using type = xfunction<functor_type, result_type, const_xclosure_t<E>...>;
+            using expression_tag = xexpression_tag_t<E...>;
+            using functor_type = build_functor_type_t<expression_tag, F, E...>;
+            using type = select_xfunction_expression_t<expression_tag, functor_type, const_xclosure_t<E>...>;
             return type(functor_type(), std::forward<E>(e)...);
         }
 
         template <template <class...> class F, class... E>
         struct xfunction_type
         {
-            using type = xfunction<F<common_value_type_t<std::decay_t<E>...>>,
-                                   typename F<common_value_type_t<std::decay_t<E>...>>::result_type,
-                                   const_xclosure_t<E>...>;
+            using expression_tag = xexpression_tag_t<E...>;
+            using functor_type = build_functor_type_t<expression_tag, F, E...>;
+            using type = select_xfunction_expression_t<expression_tag,
+                                                       functor_type,
+                                                       const_xclosure_t<E>...>;
         };
 
         // On MSVC, the second argument of enable_if_t is always evaluated, even if the condition is false.
@@ -97,6 +248,13 @@ namespace xt
                                                            xfunction_type<F, E...>>::type;
     }
 
+#undef UNARY_OPERATOR_FUNCTOR
+#undef UNARY_BOOL_OPERATOR_FUNCTOR
+#undef UNARY_OPERATOR_FUNCTOR_IMPL
+#undef BINARY_OPERATOR_FUNCTOR
+#undef BINARY_BOOL_OPERATOR_FUNCTOR
+#undef BINARY_OPERATOR_FUNCTOR_IMPL
+
     /*************
      * operators *
      *************/
@@ -132,9 +290,9 @@ namespace xt
     */
     template <class E>
     inline auto operator-(E&& e) noexcept
-        -> detail::xfunction_type_t<std::negate, E>
+        -> detail::xfunction_type_t<detail::negate, E>
     {
-        return detail::make_xfunction<std::negate>(std::forward<E>(e));
+        return detail::make_xfunction<detail::negate>(std::forward<E>(e));
     }
 
     /**
@@ -149,9 +307,9 @@ namespace xt
     */
     template <class E1, class E2>
     inline auto operator+(E1&& e1, E2&& e2) noexcept
-        -> detail::xfunction_type_t<std::plus, E1, E2>
+        -> detail::xfunction_type_t<detail::plus, E1, E2>
     {
-        return detail::make_xfunction<std::plus>(std::forward<E1>(e1), std::forward<E2>(e2));
+        return detail::make_xfunction<detail::plus>(std::forward<E1>(e1), std::forward<E2>(e2));
     }
 
     /**
@@ -166,9 +324,9 @@ namespace xt
     */
     template <class E1, class E2>
     inline auto operator-(E1&& e1, E2&& e2) noexcept
-        -> detail::xfunction_type_t<std::minus, E1, E2>
+        -> detail::xfunction_type_t<detail::minus, E1, E2>
     {
-        return detail::make_xfunction<std::minus>(std::forward<E1>(e1), std::forward<E2>(e2));
+        return detail::make_xfunction<detail::minus>(std::forward<E1>(e1), std::forward<E2>(e2));
     }
 
     /**
@@ -183,9 +341,9 @@ namespace xt
     */
     template <class E1, class E2>
     inline auto operator*(E1&& e1, E2&& e2) noexcept
-        -> detail::xfunction_type_t<std::multiplies, E1, E2>
+        -> detail::xfunction_type_t<detail::multiplies, E1, E2>
     {
-        return detail::make_xfunction<std::multiplies>(std::forward<E1>(e1), std::forward<E2>(e2));
+        return detail::make_xfunction<detail::multiplies>(std::forward<E1>(e1), std::forward<E2>(e2));
     }
 
     /**
@@ -200,9 +358,9 @@ namespace xt
     */
     template <class E1, class E2>
     inline auto operator/(E1&& e1, E2&& e2) noexcept
-        -> detail::xfunction_type_t<std::divides, E1, E2>
+        -> detail::xfunction_type_t<detail::divides, E1, E2>
     {
-        return detail::make_xfunction<std::divides>(std::forward<E1>(e1), std::forward<E2>(e2));
+        return detail::make_xfunction<detail::divides>(std::forward<E1>(e1), std::forward<E2>(e2));
     }
 
     /**
@@ -221,9 +379,9 @@ namespace xt
      */
     template <class E1, class E2>
     inline auto operator||(E1&& e1, E2&& e2) noexcept
-        -> detail::xfunction_type_t<std::logical_or, E1, E2>
+        -> detail::xfunction_type_t<detail::logical_or, E1, E2>
     {
-        return detail::make_xfunction<std::logical_or>(std::forward<E1>(e1), std::forward<E2>(e2));
+        return detail::make_xfunction<detail::logical_or>(std::forward<E1>(e1), std::forward<E2>(e2));
     }
 
     /**
@@ -238,9 +396,9 @@ namespace xt
     */
     template <class E1, class E2>
     inline auto operator&&(E1&& e1, E2&& e2) noexcept
-        -> detail::xfunction_type_t<std::logical_and, E1, E2>
+        -> detail::xfunction_type_t<detail::logical_and, E1, E2>
     {
-        return detail::make_xfunction<std::logical_and>(std::forward<E1>(e1), std::forward<E2>(e2));
+        return detail::make_xfunction<detail::logical_and>(std::forward<E1>(e1), std::forward<E2>(e2));
     }
 
     /**
@@ -254,9 +412,80 @@ namespace xt
     */
     template <class E>
     inline auto operator!(E&& e) noexcept
-        -> detail::xfunction_type_t<std::logical_not, E>
+        -> detail::xfunction_type_t<detail::logical_not, E>
     {
-        return detail::make_xfunction<std::logical_not>(std::forward<E>(e));
+        return detail::make_xfunction<detail::logical_not>(std::forward<E>(e));
+    }
+
+    /**
+     * @defgroup bitwise_operators Bitwise operators
+     */
+
+    /**
+     * @ingroup bitwise_operators
+     * @brief Bitwise and
+     *
+     * Returns an \ref xfunction for the element-wise bitwise and
+     * of \a e1 and \a e2.
+     * @param e1 an \ref xexpression or a scalar
+     * @param e2 an \ref xexpression or a scalar
+     * @return an \ref xfunction
+     */
+    template <class E1, class E2>
+    inline auto operator&(E1&& e1, E2&& e2) noexcept
+        -> detail::xfunction_type_t<detail::bitwise_and, E1, E2>
+    {
+        return detail::make_xfunction<detail::bitwise_and>(std::forward<E1>(e1), std::forward<E2>(e2));
+    }
+
+    /**
+     * @ingroup bitwise_operators
+     * @brief Bitwise or
+     *
+     * Returns an \ref xfunction for the element-wise bitwise or
+     * of \a e1 and \a e2.
+     * @param e1 an \ref xexpression or a scalar
+     * @param e2 an \ref xexpression or a scalar
+     * @return an \ref xfunction
+     */
+    template <class E1, class E2>
+    inline auto operator|(E1&& e1, E2&& e2) noexcept
+        -> detail::xfunction_type_t<detail::bitwise_or, E1, E2>
+    {
+        return detail::make_xfunction<detail::bitwise_or>(std::forward<E1>(e1), std::forward<E2>(e2));
+    }
+
+    /**
+     * @ingroup bitwise_operators
+     * @brief Bitwise xor
+     *
+     * Returns an \ref xfunction for the element-wise bitwise xor
+     * of \a e1 and \a e2.
+     * @param e1 an \ref xexpression or a scalar
+     * @param e2 an \ref xexpression or a scalar
+     * @return an \ref xfunction
+     */
+    template <class E1, class E2>
+    inline auto operator^(E1&& e1, E2&& e2) noexcept
+        -> detail::xfunction_type_t<detail::bitwise_xor, E1, E2>
+    {
+        return detail::make_xfunction<detail::bitwise_xor>(std::forward<E1>(e1), std::forward<E2>(e2));
+    }
+
+    /**
+     * @ingroup bitwise_operators
+     * @brief Bitwise not
+     *
+     * Returns an \ref xfunction for the element-wise bitwise not
+     * of \a e.
+     * @param e an \ref xexpression
+     * @return an \ref xfunction
+     */
+    template <class E>
+    inline auto operator~(E&& e) noexcept
+        -> detail::xfunction_type_t<detail::bitwise_not, E>
+    {
+        return detail::make_xfunction<detail::bitwise_not>(std::forward<E>(e));
     }
 
     /**
@@ -275,9 +504,9 @@ namespace xt
      */
     template <class E1, class E2>
     inline auto operator<(E1&& e1, E2&& e2) noexcept
-        -> detail::xfunction_type_t<std::less, E1, E2>
+        -> detail::xfunction_type_t<detail::less, E1, E2>
     {
-        return detail::make_xfunction<std::less>(std::forward<E1>(e1), std::forward<E2>(e2));
+        return detail::make_xfunction<detail::less>(std::forward<E1>(e1), std::forward<E2>(e2));
     }
 
     /**
@@ -292,9 +521,9 @@ namespace xt
     */
     template <class E1, class E2>
     inline auto operator<=(E1&& e1, E2&& e2) noexcept
-        -> detail::xfunction_type_t<std::less_equal, E1, E2>
+        -> detail::xfunction_type_t<detail::less_equal, E1, E2>
     {
-        return detail::make_xfunction<std::less_equal>(std::forward<E1>(e1), std::forward<E2>(e2));
+        return detail::make_xfunction<detail::less_equal>(std::forward<E1>(e1), std::forward<E2>(e2));
     }
 
     /**
@@ -309,9 +538,9 @@ namespace xt
     */
     template <class E1, class E2>
     inline auto operator>(E1&& e1, E2&& e2) noexcept
-        -> detail::xfunction_type_t<std::greater, E1, E2>
+        -> detail::xfunction_type_t<detail::greater, E1, E2>
     {
-        return detail::make_xfunction<std::greater>(std::forward<E1>(e1), std::forward<E2>(e2));
+        return detail::make_xfunction<detail::greater>(std::forward<E1>(e1), std::forward<E2>(e2));
     }
 
     /**
@@ -326,9 +555,9 @@ namespace xt
     */
     template <class E1, class E2>
     inline auto operator>=(E1&& e1, E2&& e2) noexcept
-        -> detail::xfunction_type_t<std::greater_equal, E1, E2>
+        -> detail::xfunction_type_t<detail::greater_equal, E1, E2>
     {
-        return detail::make_xfunction<std::greater_equal>(std::forward<E1>(e1), std::forward<E2>(e2));
+        return detail::make_xfunction<detail::greater_equal>(std::forward<E1>(e1), std::forward<E2>(e2));
     }
 
     /**
@@ -348,9 +577,9 @@ namespace xt
         const E1& de1 = e1.derived_cast();
         const E2& de2 = e2.derived_cast();
         bool res = de1.dimension() == de2.dimension() && std::equal(de1.shape().begin(), de1.shape().end(), de2.shape().begin());
-        auto iter1 = de1.xbegin();
-        auto iter2 = de2.xbegin();
-        auto iter_end = de1.xend();
+        auto iter1 = de1.begin();
+        auto iter2 = de2.begin();
+        auto iter_end = de1.end();
         while (res && iter1 != iter_end)
         {
             res = (*iter1++ == *iter2++);
@@ -387,9 +616,9 @@ namespace xt
     */
     template <class E1, class E2>
     inline auto equal(E1&& e1, E2&& e2) noexcept
-        -> detail::xfunction_type_t<std::equal_to, E1, E2>
+        -> detail::xfunction_type_t<detail::equal_to, E1, E2>
     {
-        return detail::make_xfunction<std::equal_to>(std::forward<E1>(e1), std::forward<E2>(e2));
+        return detail::make_xfunction<detail::equal_to>(std::forward<E1>(e1), std::forward<E2>(e2));
     }
 
     /**
@@ -404,9 +633,9 @@ namespace xt
     */
     template <class E1, class E2>
     inline auto not_equal(E1&& e1, E2&& e2) noexcept
-        -> detail::xfunction_type_t<std::not_equal_to, E1, E2>
+        -> detail::xfunction_type_t<detail::not_equal_to, E1, E2>
     {
-        return detail::make_xfunction<std::not_equal_to>(std::forward<E1>(e1), std::forward<E2>(e2));
+        return detail::make_xfunction<detail::not_equal_to>(std::forward<E1>(e1), std::forward<E2>(e2));
     }
 
     /**
@@ -428,124 +657,10 @@ namespace xt
         return detail::make_xfunction<detail::conditional_ternary>(std::forward<E1>(e1), std::forward<E2>(e2), std::forward<E3>(e3));
     }
 
-    /**
-    * @ingroup logical_operators
-    * @brief Elementwise maximum
-    *
-    * Returns an \ref xfunction for the element-wise
-    * maximum between e1 and e2.
-    * @param e1 an \ref xexpression
-    * @param e2 an \ref xexpression
-    * @return an \ref xfunction
-    */
-    template <class E1, class E2>
-    inline auto maximum(E1&& e1, E2&& e2) noexcept
-        -> detail::xfunction_type_t<detail::maximum, E1, E2>
-    {
-        return detail::make_xfunction<detail::maximum>(std::forward<E1>(e1), std::forward<E2>(e2));
-    }
-
-    /**
-    * @ingroup logical_operators
-    * @brief Elementwise minimum
-    *
-    * Returns an \ref xfunction for the element-wise
-    * minimum between e1 and e2.
-    * @param e1 an \ref xexpression
-    * @param e2 an \ref xexpression
-    * @return an \ref xfunction
-    */
-    template <class E1, class E2>
-    inline auto minimum(E1&& e1, E2&& e2) noexcept
-        -> detail::xfunction_type_t<detail::minimum, E1, E2>
-    {
-        return detail::make_xfunction<detail::minimum>(std::forward<E1>(e1), std::forward<E2>(e2));
-    }
-
-    /**
-     * @ingroup logical_operators
-     * @brief Maximum element along given axis.
-     *
-     * Returns an \ref xreducer for the maximum of elements over given
-     * \em axes.
-     * @param e an \ref xexpression
-     * @param axes the axes along which the maximum is found (optional)
-     * @return an \ref xreducer
-     */
-    template <class E, class X>
-    inline auto amax(E&& e, X&& axes) noexcept
-    {
-        using functor_type = detail::maximum<typename std::decay_t<E>::value_type>;
-        return reduce(functor_type(), std::forward<E>(e), std::forward<X>(axes));
-    }
-
-    template <class E>
-    inline auto amax(E&& e) noexcept
-    {
-        using functor_type = detail::maximum<typename std::decay_t<E>::value_type>;
-        return reduce(functor_type(), std::forward<E>(e));
-    }
-
-#ifdef X_OLD_CLANG
-    template <class E, class I>
-    inline auto amax(E&& e, std::initializer_list<I> axes) noexcept
-    {
-        using functor_type = detail::maximum<typename std::decay_t<E>::value_type>;
-        return reduce(functor_type(), std::forward<E>(e), axes);
-    }
-#else
-    template <class E, class I, std::size_t N>
-    inline auto amax(E&& e, const I (&axes)[N]) noexcept
-    {
-        using functor_type = detail::maximum<typename std::decay_t<E>::value_type>;
-        return reduce(functor_type(), std::forward<E>(e), axes);
-    }
-#endif
-
-    /**
-     * @ingroup logical_operators
-     * @brief Minimum element along given axis.
-     *
-     * Returns an \ref xreducer for the minimum of elements over given
-     * \em axes.
-     * @param e an \ref xexpression
-     * @param axes the axes along which the minimum is found (optional)
-     * @return an \ref xreducer
-     */
-    template <class E, class X>
-    inline auto amin(E&& e, X&& axes) noexcept
-    {
-        using functor_type = detail::minimum<typename std::decay_t<E>::value_type>;
-        return reduce(functor_type(), std::forward<E>(e), std::forward<X>(axes));
-    }
-
-    template <class E>
-    inline auto amin(E&& e) noexcept
-    {
-        using functor_type = detail::minimum<typename std::decay_t<E>::value_type>;
-        return reduce(functor_type(), std::forward<E>(e));
-    }
-
-#ifdef X_OLD_CLANG
-    template <class E, class I>
-    inline auto amin(E&& e, std::initializer_list<I> axes) noexcept
-    {
-        using functor_type = detail::minimum<typename std::decay_t<E>::value_type>;
-        return reduce(functor_type(), std::forward<E>(e), axes);
-    }
-#else
-    template <class E, class I, std::size_t N>
-    inline auto amin(E&& e, const I (&axes)[N]) noexcept
-    {
-        using functor_type = detail::minimum<typename std::decay_t<E>::value_type>;
-        return reduce(functor_type(), std::forward<E>(e), axes);
-    }
-#endif
-
     /**
      * @ingroup logical_operators
      * @brief return vector of indices where T is not zero
-     * 
+     *
      * @param arr input array
      * @return vector of \a index_types where arr is not equal to zero
      */
@@ -557,13 +672,14 @@ namespace xt
         using index_type = xindex_type_t<typename T::shape_type>;
         using size_type = typename T::size_type;
 
-        index_type idx(arr.dimension(), 0);
+        auto idx = xtl::make_sequence<index_type>(arr.dimension(), 0);
         std::vector<index_type> indices;
 
         auto next_idx = [&shape](index_type& idx)
         {
-            for (int i = int(shape.size() - 1); i >= 0; --i)
+            for (size_type j = shape.size(); j > 0; --j)
             {
+                size_type i = j - 1;
                 if (idx[i] >= shape[i] - 1)
                 {
                     idx[i] = 0;
@@ -581,7 +697,7 @@ namespace xt
         size_type total_size = compute_size(shape);
         for (size_type i = 0; i < total_size; i++, next_idx(idx))
         {
-            if (arr[idx])
+            if (arr.element(std::begin(idx), std::end(idx)))
             {
                 indices.push_back(idx);
             }
@@ -593,7 +709,7 @@ namespace xt
      * @ingroup logical_operators
      * @brief return vector of indices where condition is true
      *        (equivalent to \a nonzero(condition))
-     * 
+     *
      * @param condition input array
      * @return vector of \a index_types where condition is not equal to zero
      */
@@ -624,7 +740,7 @@ namespace xt
         }
         else
         {
-            return std::any_of(e.xbegin(), e.xend(),
+            return std::any_of(e.begin(), e.end(),
                                [](const typename std::decay_t<E>::value_type& el) { return el; });
         }
     }
@@ -649,10 +765,33 @@ namespace xt
         }
         else
         {
-            return std::all_of(e.xbegin(), e.xend(),
+            return std::all_of(e.begin(), e.end(),
                                [](const typename std::decay_t<E>::value_type& el) { return el; });
         }
     }
+
+    /**
+     * @defgroup casting_operators Casting operators
+     */
+    
+    /**
+     * @ingroup casting_operators
+     * @brief Element-wise ``static_cast``.
+     *
+     * Returns an \ref xfunction for the element-wise
+     * static_cast of \a e to type R.
+     *
+     * @param e an \ref xexpression or a scalar
+     * @return an \ref xfunction
+     */
+    
+    template <class R, class E>
+    inline auto cast(E&& e) noexcept
+    -> detail::xfunction_type_t<detail::cast<R>::template functor, E>
+    {
+        return detail::make_xfunction<detail::cast<R>::template functor>(std::forward<E>(e));
+    }
+
 }
 
 #endif
