@@ -6,8 +6,8 @@
 * The full license is in the file LICENSE, distributed with this software. *
 ****************************************************************************/
 
-#ifndef XREDUCER_HPP
-#define XREDUCER_HPP
+#ifndef XTENSOR_REDUCER_HPP
+#define XTENSOR_REDUCER_HPP
 
 #include <algorithm>
 #include <cstddef>
@@ -16,10 +16,13 @@
 #include <stdexcept>
 #include <type_traits>
 #include <utility>
+#include <tuple>
 #ifdef X_OLD_CLANG
 #include <vector>
 #endif
 
+#include "xtl/xsequence.hpp"
+
 #include "xbuilder.hpp"
 #include "xexpression.hpp"
 #include "xgenerator.hpp"
@@ -29,7 +32,6 @@
 
 namespace xt
 {
-
     /**********
      * reduce *
      **********/
@@ -55,6 +57,59 @@ namespace xt
     template <class ST, class X>
     struct xreducer_shape_type;
 
+    template <class REDUCE_FUNC, class INIT_FUNC=identity_functor, class MERGE_FUNC=REDUCE_FUNC>
+    struct xreducer_functors
+    :  public std::tuple<REDUCE_FUNC, INIT_FUNC, MERGE_FUNC>
+    {
+        using self_type = xreducer_functors<REDUCE_FUNC, INIT_FUNC, MERGE_FUNC>;
+        using base_type = std::tuple<REDUCE_FUNC, INIT_FUNC, MERGE_FUNC>;
+        using reduce_functor_type = REDUCE_FUNC;
+        using init_functor_type = INIT_FUNC;
+        using merge_functor_type = MERGE_FUNC;
+
+        xreducer_functors()
+        : base_type()
+        {}
+
+        template <class RF>
+        xreducer_functors(RF&& reduce_func)
+        : base_type(std::forward<RF>(reduce_func), INIT_FUNC(), reduce_func)
+        {}
+
+        template <class RF, class IF>
+        xreducer_functors(RF&& reduce_func, IF&& init_func)
+        : base_type(std::forward<RF>(reduce_func), std::forward<IF>(init_func), reduce_func)
+        {}
+
+        template <class RF, class IF, class MF>
+        xreducer_functors(RF&& reduce_func, IF&& init_func, MF&& merge_func)
+        : base_type(std::forward<RF>(reduce_func), std::forward<IF>(init_func), std::forward<MF>(merge_func))
+        {}
+    };
+
+    template <class RF>
+    auto make_xreducer_functor(RF && reduce_func)
+    {
+        using reducer_type = xreducer_functors<std::remove_reference_t<RF>>;
+        return reducer_type(std::forward<RF>(reduce_func));
+    }
+
+    template <class RF, class IF>
+    auto make_xreducer_functor(RF && reduce_func, IF && init_func)
+    {
+        using reducer_type = xreducer_functors<std::remove_reference_t<RF>, std::remove_reference_t<IF>>;
+        return reducer_type(std::forward<RF>(reduce_func), std::forward<IF>(init_func));
+    }
+
+    template <class RF, class IF, class MF>
+    auto make_xreducer_functor(RF && reduce_func, IF && init_func, MF && merge_func)
+    {
+        using reducer_type = xreducer_functors<std::remove_reference_t<RF>,
+                                               std::remove_reference_t<IF>,
+                                               std::remove_reference_t<MF>>;
+        return reducer_type(std::forward<RF>(reduce_func), std::forward<IF>(init_func), std::forward<MF>(merge_func));
+    }
+
     template <class F, class CT, class X>
     class xreducer;
 
@@ -68,10 +123,6 @@ namespace xt
         using inner_shape_type = typename xreducer_shape_type<typename xexpression_type::shape_type, X>::type;
         using const_stepper = xreducer_stepper<F, CT, X>;
         using stepper = const_stepper;
-        using const_iterator = xiterator<const_stepper, inner_shape_type*, DEFAULT_LAYOUT>;
-        using iterator = const_iterator;
-        using const_reverse_iterator = std::reverse_iterator<const_iterator>;
-        using reverse_iterator = std::reverse_iterator<iterator>;
     };
 
     /**
@@ -82,24 +133,31 @@ namespace xt
      * a reducing function to an \ref xexpression over the specified
      * axes.
      *
-     * @tparam F the function type
+     * @tparam F a tuple of functors (class \ref xreducer_functors or compatible)
      * @tparam CT the closure type of the \ref xexpression to reduce
      * @tparam X the list of axes
      *
+     * The reducer's result_type is deduced from the result type of function
+     * <tt>F::reduce_functor_type</tt> when called with elements of the expression @tparam CT.
+     *
      * @sa reduce
      */
     template <class F, class CT, class X>
     class xreducer : public xexpression<xreducer<F, CT, X>>,
-                     public xexpression_const_iterable<xreducer<F, CT, X>>
+                     public xconst_iterable<xreducer<F, CT, X>>
     {
     public:
 
         using self_type = xreducer<F, CT, X>;
-        using functor_type = typename std::remove_reference<F>::type;
+        using reduce_functor_type = typename F::reduce_functor_type;
+        using init_functor_type = typename F::init_functor_type;
+        using merge_functor_type = typename F::merge_functor_type;
         using xexpression_type = std::decay_t<CT>;
         using axes_type = X;
 
-        using value_type = typename xexpression_type::value_type;
+        using substepper_type = typename xexpression_type::const_stepper;
+        using value_type = std::decay_t<decltype(std::declval<reduce_functor_type>()(
+            std::declval<init_functor_type>()(**(substepper_type*)0), **(substepper_type*)0))>;
         using reference = value_type;
         using const_reference = value_type;
         using pointer = value_type*;
@@ -108,7 +166,7 @@ namespace xt
         using size_type = typename xexpression_type::size_type;
         using difference_type = typename xexpression_type::difference_type;
 
-        using iterable_base = xexpression_const_iterable<self_type>;
+        using iterable_base = xconst_iterable<self_type>;
         using inner_shape_type = typename iterable_base::inner_shape_type;
         using shape_type = inner_shape_type;
 
@@ -128,6 +186,8 @@ namespace xt
 
         template <class... Args>
         const_reference operator()(Args... args) const;
+        template <class... Args>
+        const_reference at(Args... args) const;
         const_reference operator[](const xindex& index) const;
         const_reference operator[](size_type i) const;
 
@@ -148,7 +208,9 @@ namespace xt
     private:
 
         CT m_e;
-        functor_type m_f;
+        reduce_functor_type m_reduce;
+        init_functor_type m_init;
+        merge_functor_type m_merge;
         axes_type m_axes;
         inner_shape_type m_shape;
         shape_type m_dim_mapping;
@@ -175,7 +237,7 @@ namespace xt
     template <class F, class E, class X>
     inline auto reduce(F&& f, E&& e, X&& axes) noexcept
     {
-        using reducer_type = xreducer<F, const_xclosure_t<E>, const_closure_t<X>>;
+        using reducer_type = xreducer<F, const_xclosure_t<E>, xtl::const_closure_type_t<X>>;
         return reducer_type(std::forward<F>(f), std::forward<E>(e), std::forward<X>(axes));
     }
 
@@ -194,7 +256,7 @@ namespace xt
     {
         using axes_type = std::vector<typename std::decay_t<E>::size_type>;
         using reducer_type = xreducer<F, const_xclosure_t<E>, axes_type>;
-        return reducer_type(std::forward<F>(f), std::forward<E>(e), forward_sequence<axes_type>(axes));
+        return reducer_type(std::forward<F>(f), std::forward<E>(e), xtl::forward_sequence<axes_type>(axes));
     }
 #else
     template <class F, class E, class I, std::size_t N>
@@ -202,7 +264,7 @@ namespace xt
     {
         using axes_type = std::array<typename std::decay_t<E>::size_type, N>;
         using reducer_type = xreducer<F, const_xclosure_t<E>, axes_type>;
-        return reducer_type(std::forward<F>(f), std::forward<E>(e), forward_sequence<axes_type>(axes));
+        return reducer_type(std::forward<F>(f), std::forward<E>(e), xtl::forward_sequence<axes_type>(axes));
     }
 #endif
 
@@ -289,6 +351,7 @@ namespace xt
                                    OutputIt d_first, OutputIt map_first)
         {
             using difference_type = typename std::iterator_traits<InputIt>::difference_type;
+            using value_type = typename std::iterator_traits<OutputIt>::value_type;
             InputIt iter = first;
             while (iter != last && e_first != e_last)
             {
@@ -296,7 +359,7 @@ namespace xt
                 if (diff != difference_type(*e_first))
                 {
                     *d_first++ = *iter++;
-                    *map_first++ = diff;
+                    *map_first++ = value_type(diff);
                 }
                 else
                 {
@@ -330,9 +393,13 @@ namespace xt
     template <class F, class CT, class X>
     template <class Func, class CTA, class AX>
     inline xreducer<F, CT, X>::xreducer(Func&& func, CTA&& e, AX&& axes)
-        : m_e(std::forward<CTA>(e)), m_f(std::forward<Func>(func)), m_axes(std::forward<AX>(axes)),
-          m_shape(make_sequence<shape_type>(m_e.dimension() - m_axes.size(), 0)),
-          m_dim_mapping(make_sequence<shape_type>(m_e.dimension() - m_axes.size(), 0))
+        : m_e(std::forward<CTA>(e))
+        , m_reduce(std::get<0>(func))
+        , m_init(std::get<1>(func))
+        , m_merge(std::get<2>(func))
+        , m_axes(std::forward<AX>(axes))
+        , m_shape(xtl::make_sequence<shape_type>(m_e.dimension() - m_axes.size(), 0))
+        , m_dim_mapping(xtl::make_sequence<shape_type>(m_e.dimension() - m_axes.size(), 0))
     {
         if (!std::is_sorted(m_axes.cbegin(), m_axes.cend()))
         {
@@ -401,6 +468,22 @@ namespace xt
         return element(arg_array.cbegin(), arg_array.cend());
     }
 
+    /**
+     * Returns a constant reference to the element at the specified position in the expression,
+     * after dimension and bounds checking.
+     * @param args a list of indices specifying the position in the function. Indices
+     * must be unsigned integers, the number of indices should be equal to the number of dimensions
+     * of the expression.
+     * @exception std::out_of_range if the number of argument is greater than the number of dimensions
+     * or if indices are out of bounds.
+     */
+    template <class F, class CT, class X>
+    template <class... Args>
+    inline auto xreducer<F, CT, X>::at(Args... args) const -> const_reference
+    {
+        check_access(shape(), static_cast<size_type>(args)...);
+        return this->operator()(args...);
+    }
     /**
      * Returns a constant reference to the element at the specified position in the reducer.
      * @param index a sequence of indices specifying the position in the reducer. Indices
@@ -423,7 +506,7 @@ namespace xt
      * Returns a constant reference to the element at the specified position in the reducer.
      * @param first iterator starting the sequence of indices
      * @param last iterator ending the sequence of indices
-     * The number of indices in the squence should be equal to or greater
+     * The number of indices in the sequence should be equal to or greater
      * than the number of dimensions of the reducer.
      */
     template <class F, class CT, class X>
@@ -511,28 +594,36 @@ namespace xt
     inline void xreducer_stepper<F, CT, X>::step(size_type dim, size_type n)
     {
         if (dim >= m_offset)
+        {
             m_stepper.step(get_dim(dim), n);
+        }
     }
 
     template <class F, class CT, class X>
     inline void xreducer_stepper<F, CT, X>::step_back(size_type dim, size_type n)
     {
         if (dim >= m_offset)
+        {
             m_stepper.step_back(get_dim(dim), n);
+        }
     }
 
     template <class F, class CT, class X>
     inline void xreducer_stepper<F, CT, X>::reset(size_type dim)
     {
         if (dim >= m_offset)
+        {
             m_stepper.reset(get_dim(dim));
+        }
     }
 
     template <class F, class CT, class X>
     inline void xreducer_stepper<F, CT, X>::reset_back(size_type dim)
     {
         if (dim >= m_offset)
+        {
             m_stepper.reset_back(get_dim(dim));
+        }
     }
 
     template <class F, class CT, class X>
@@ -565,16 +656,16 @@ namespace xt
             for (size_type i = 1; i != size; ++i)
             {
                 m_stepper.step(index);
-                res = m_reducer.m_f(res, aggregate(dim + 1));
+                res = m_reducer.m_merge(res, aggregate(dim + 1));
             }
         }
         else
         {
-            res = *m_stepper;
+            res = m_reducer.m_init(*m_stepper);
             for (size_type i = 1; i != size; ++i)
             {
                 m_stepper.step(index);
-                res = m_reducer.m_f(res, *m_stepper);
+                res = m_reducer.m_reduce(res, *m_stepper);
             }
         }
         m_stepper.reset(index);
