@@ -6,11 +6,12 @@
 * The full license is in the file LICENSE, distributed with this software. *
 ****************************************************************************/
 
-#ifndef XUTILS_HPP
-#define XUTILS_HPP
+#ifndef XTENSOR_UTILS_HPP
+#define XTENSOR_UTILS_HPP
 
 #include <algorithm>
 #include <array>
+#include <cmath>
 #include <complex>
 #include <cstddef>
 #include <initializer_list>
@@ -19,6 +20,8 @@
 #include <utility>
 #include <vector>
 
+#include "xtl/xtype_traits.hpp"
+
 #include "xtensor_config.hpp"
 
 namespace xt
@@ -36,18 +39,6 @@ namespace xt
     template <class F, class R, class... T>
     R accumulate(F&& f, R init, const std::tuple<T...>& t) noexcept(noexcept(std::declval<F>()));
 
-    template <class... T>
-    struct or_;
-
-    template <class... T>
-    struct and_;
-
-    template <bool... B>
-    struct or_c;
-
-    template <bool... B>
-    struct and_c;
-
     template <std::size_t I, class... Args>
     constexpr decltype(auto) argument(Args&&... args) noexcept;
 
@@ -75,19 +66,16 @@ namespace xt
     template <class T, std::size_t N>
     bool resize_container(std::array<T, N>& a, typename std::array<T, N>::size_type size);
 
-    template <class S>
-    S make_sequence(typename S::size_type size, typename S::value_type v);
+    // gcc 4.9 is affected by C++14 defect CGW 1558
+    // see http://open-std.org/JTC1/SC22/WG21/docs/cwg_defects.html#1558
+    template <class... T>
+    struct make_void
+    {
+        using type = void;
+    };
 
-    template <class R, class A>
-    decltype(auto) forward_sequence(A&& s);
-
-    // equivalent to std::size(c) in c++17
-    template <class C>
-    constexpr auto sequence_size(const C& c) -> decltype(c.size());
-
-    // equivalent to std::size(a) in c++17
-    template <class T, std::size_t N>
-    constexpr std::size_t sequence_size(const T (&a)[N]);
+    template <class... T>
+    using void_t = typename make_void<T...>::type;
 
     /*******************************
      * remove_class implementation *
@@ -168,50 +156,6 @@ namespace xt
         return detail::accumulate_impl<0, F, R, T...>(f, init, t);
     }
 
-    /**********************
-     * or_ implementation *
-     **********************/
-
-    template <>
-    struct or_<> : std::integral_constant<bool, false>
-    {
-    };
-
-    template <class T, class... Ts>
-    struct or_<T, Ts...>
-        : std::integral_constant<bool, T::value || or_<Ts...>::value>
-    {
-    };
-
-    /***********************
-     * and_ implementation *
-     ***********************/
-
-    template <>
-    struct and_<> : std::integral_constant<bool, true>
-    {
-    };
-
-    template <class T, class... Ts>
-    struct and_<T, Ts...>
-        : std::integral_constant<bool, T::value && and_<Ts...>::value>
-    {
-    };
-
-    /**********************************
-     * or_c and and_c implementations *
-     **********************************/
-
-    template <bool... B>
-    struct or_c : or_<std::integral_constant<bool, B>...>
-    {
-    };
-
-    template <bool... B>
-    struct and_c : and_<std::integral_constant<bool, B>...>
-    {
-    };
-
     /***************************
      * argument implementation *
      ***************************/
@@ -255,7 +199,7 @@ namespace xt
         template <class R, class F, std::size_t I, class... S>
         R apply_one(F&& func, const std::tuple<S...>& s) noexcept(noexcept(std::declval<F>()))
         {
-            return func(std::get<I>(s));
+            return static_cast<R>(func(std::get<I>(s)));
         }
 
         template <class R, class F, std::size_t... I, class... S>
@@ -440,104 +384,6 @@ namespace xt
         return size == N;
     }
 
-    /********************************
-     * make_sequence implementation *
-     ********************************/
-
-    namespace detail
-    {
-        template <class S>
-        struct sequence_builder
-        {
-            using value_type = typename S::value_type;
-            using size_type = typename S::size_type;
-
-            inline static S make(size_type size, value_type v)
-            {
-                return S(size, v);
-            }
-        };
-
-        template <class T, std::size_t N>
-        struct sequence_builder<std::array<T, N>>
-        {
-            using sequence_type = std::array<T, N>;
-            using value_type = typename sequence_type::value_type;
-            using size_type = typename sequence_type::size_type;
-
-            inline static sequence_type make(size_type /*size*/, value_type v)
-            {
-                sequence_type s;
-                s.fill(v);
-                return s;
-            }
-        };
-    }
-
-    template <class S>
-    inline S make_sequence(typename S::size_type size, typename S::value_type v)
-    {
-        return detail::sequence_builder<S>::make(size, v);
-    }
-
-    /***********************************
-     * forward_sequence implementation *
-     ***********************************/
-
-    namespace detail
-    {
-        template <class R, class A, class E = void>
-        struct sequence_forwarder
-        {
-            template <class T>
-            static inline R forward(const T& r)
-            {
-                return R(std::begin(r), std::end(r));
-            }
-        };
-
-        template <class I, std::size_t L, class A>
-        struct sequence_forwarder<std::array<I, L>, A,
-                                  std::enable_if_t<!std::is_same<std::array<I, L>, A>::value>>
-        {
-            using R = std::array<I, L>;
-
-            template <class T>
-            static inline R forward(const T& r)
-            {
-                R ret;
-                std::copy(std::begin(r), std::end(r), std::begin(ret));
-                return ret;
-            }
-        };
-
-        template <class R>
-        struct sequence_forwarder<R, R>
-        {
-            template <class T>
-            static inline T&& forward(typename std::remove_reference<T>::type& t) noexcept
-            {
-                return static_cast<T&&>(t);
-            }
-
-            template <class T>
-            static inline T&& forward(typename std::remove_reference<T>::type&& t) noexcept
-            {
-                return static_cast<T&&>(t);
-            }
-        };
-    }
-
-    template <class R, class A>
-    inline decltype(auto) forward_sequence(A&& s)
-    {
-        using forwarder = detail::sequence_forwarder<
-            std::decay_t<R>,
-            std::remove_cv_t<std::remove_reference_t<A>>
-        >;
-        return forwarder::template forward<A>(s);
-    }
-
     /*************************************
      * promote_shape and promote_strides *
      *************************************/
@@ -579,7 +425,7 @@ namespace xt
         };
 
         template <class... S>
-        using only_array = and_<is_array<S>...>;
+        using only_array = xtl::conjunction<is_array<S>...>;
 
         // The promote_index meta-function returns std::vector<promoted_value_type> in the
         // general case and an array of the promoted value type and maximal size if all
@@ -619,67 +465,6 @@ namespace xt
     template <class... S>
     using promote_strides_t = typename detail::promote_index<S...>::type;
 
-
-    /**************************
-     * closure implementation *
-     **************************/
-
-    template <class S>
-    struct closure
-    {
-        using underlying_type = std::conditional_t<std::is_const<std::remove_reference_t<S>>::value,
-                                                   const std::decay_t<S>,
-                                                   std::decay_t<S>>;
-        using type = typename std::conditional<std::is_lvalue_reference<S>::value,
-                                               underlying_type&,
-                                               underlying_type>::type;
-    };
-
-    template <class S>
-    using closure_t = typename closure<S>::type;
-
-    template <class S>
-    struct const_closure
-    {
-        using underlying_type = const std::decay_t<S>;
-        using type = typename std::conditional<std::is_lvalue_reference<S>::value,
-                                               underlying_type&,
-                                               underlying_type>::type;
-    };
-
-    template <class S>
-    using const_closure_t = typename const_closure<S>::type;
-
-    /******************************
-     * ptr_closure implementation *
-     ******************************/
-
-    template <class S>
-    struct ptr_closure
-    {
-        using underlying_type = std::conditional_t<std::is_const<std::remove_reference_t<S>>::value,
-                                                   const std::decay_t<S>,
-                                                   std::decay_t<S>>;
-        using type = std::conditional_t<std::is_lvalue_reference<S>::value,
-                                        underlying_type*,
-                                        underlying_type>;
-    };
-
-    template <class S>
-    using ptr_closure_t = typename ptr_closure<S>::type;
-
-    template <class S>
-    struct const_ptr_closure
-    {
-        using underlying_type = const std::decay_t<S>;
-        using type = std::conditional_t<std::is_lvalue_reference<S>::value,
-                                        underlying_type*,
-                                        underlying_type>;
-    };
-
-    template <class S>
-    using const_ptr_closure_t = typename const_ptr_closure<S>::type;
-
     /***************************
      * apply_cv implementation *
      ***************************/
@@ -745,121 +530,6 @@ namespace xt
     template <class T, class U>
     using apply_cv_t = typename apply_cv<T, U>::type;
 
-    /*****************************
-     * is_complex implementation *
-     *****************************/
-
-    namespace detail
-    {
-        template <typename T>
-        struct is_complex : public std::false_type
-        {
-        };
-
-        template <typename T>
-        struct is_complex<std::complex<T>> : public std::true_type
-        {
-        };
-    }
-
-    template <class T>
-    struct is_complex
-    {
-        static constexpr bool value = detail::is_complex<std::decay_t<T>>::value;
-    };
-
-    /*************************************
-     * complex_value_type implementation *
-     *************************************/
-
-    template <typename T>
-    struct complex_value_type
-    {
-        using type = T;
-    };
-
-    template <typename T>
-    struct complex_value_type<std::complex<T>>
-    {
-        using type = T;
-    };
-
-    template <class T>
-    using complex_value_type_t = typename complex_value_type<T>::type;
-
-    /*********************************
-     * forward_offset implementation *
-     *********************************/
-
-    namespace detail
-    {
-
-        template <class T, class M>
-        struct forward_type
-        {
-            using type = apply_cv_t<T, M>&&;
-        };
-
-        template <class T, class M>
-        struct forward_type<T&, M>
-        {
-            using type = apply_cv_t<T, M>&;
-        };
-
-        template <class T, class M>
-        using forward_type_t = typename forward_type<T, M>::type;
-    }
-
-    template <class M, std::size_t I, class T>
-    constexpr detail::forward_type_t<T, M> forward_offset(T&& v) noexcept
-    {
-        using forward_type = detail::forward_type_t<T, M>;
-        using cv_value_type = std::remove_reference_t<forward_type>;
-        using byte_type = apply_cv_t<std::remove_reference_t<T>, char>;
-
-        return static_cast<forward_type>(
-            *reinterpret_cast<cv_value_type*>(
-                reinterpret_cast<byte_type*>(&v) + I
-            )
-        );
-    }
-
-    /**********************************************
-     * forward_real & forward_imag implementation *
-     **********************************************/
-
-    // forward_real
-
-    template <class T>
-    auto forward_real(T&& v)
-        -> std::enable_if_t<!is_complex<T>::value, detail::forward_type_t<T, T>>  // real case -> forward
-    {
-        return static_cast<detail::forward_type_t<T, T>>(v);
-    }
-
-    template <class T>
-    auto forward_real(T&& v)
-        -> std::enable_if_t<is_complex<T>::value, detail::forward_type_t<T, typename std::decay_t<T>::value_type>>  // complex case -> forward the real part
-    {
-        return forward_offset<typename std::decay_t<T>::value_type, 0>(v);
-    }
-
-    // forward_imag
-
-    template <class T>
-    auto forward_imag(T &&)
-        -> std::enable_if_t<!is_complex<T>::value, std::decay_t<T>>  // real case -> always return 0 by value
-    {
-        return 0;
-    }
-
-    template <class T>
-    auto forward_imag(T&& v)
-        -> std::enable_if_t<is_complex<T>::value, detail::forward_type_t<T, typename std::decay_t<T>::value_type>>  // complex case -> forwards the imaginary part
-    {
-        using real_type = typename std::decay_t<T>::value_type;
-        return forward_offset<real_type, sizeof(real_type)>(v);
-    }
 
     /**************************
     * to_array implementation *
@@ -902,13 +572,13 @@ namespace xt
      * has_raw_data_interface implementation *
      *****************************************/
 
-    template <typename T>
+    template <class T>
     class has_raw_data_interface
     {
-        template <typename C>
+        template <class C>
         static std::true_type test(decltype(std::declval<C>().raw_data_offset()));
 
-        template <typename C>
+        template <class C>
         static std::false_type test(...);
 
     public:
@@ -916,80 +586,434 @@ namespace xt
         constexpr static bool value = decltype(test<T>(std::size_t(0)))::value == true;
     };
 
+    /******************
+     * enable_if_type *
+     ******************/
+
+    template <class T>
+    struct enable_if_type
+    {
+        using type = void;
+    };
+
     /*****************************
-     * is_complete implemenation * 
+     * is_complete implemenation *
      *****************************/
 
     namespace detail
     {
-        template <typename T>
-        struct is_complete_impl
+        template <class T>
+        class is_complete_impl
         {
-            template <typename U>
+            template <class U>
             static auto test(U*)  -> std::integral_constant<bool, sizeof(U) == sizeof(U)>;
+
             static auto test(...) -> std::false_type;
+
+        public:
+
             using type = decltype(test((T*)0));
         };
     }
 
-    template <typename T>
+    template <class T>
     struct is_complete : detail::is_complete_impl<T>::type {};
 
-    /********************************************
-     * xtrivial_default_construct implemenation * 
-     ********************************************/
+    /*************
+     * static_if *
+     *************/
 
-#if defined(__clang__)
-#if !(defined(__APPLE__))
-// CLANG && LINUX
-#pragma clang diagnostic push
-#pragma clang diagnostic ignored "-Wdeprecated-declarations"
-
-}
-namespace std { template <class T> struct is_trivially_default_constructible; }
-namespace xt
-{
-
-    namespace detail
+    struct identity_functor
     {
-        template <bool C, class T>
-        struct xtrivial_default_construct_impl;
-
         template <class T>
-        struct xtrivial_default_construct_impl<true, T> : std::is_trivially_default_constructible<T> {};
+        T&& operator()(T&& x) const
+        {
+            return std::forward<T>(x);
+        }
+    };
 
-        template <class T>
-        struct xtrivial_default_construct_impl<false, T> : std::has_trivial_default_constructor<T> {};
+    template <class TF, class FF>
+    auto static_if(std::true_type, const TF& tf, const FF&)
+    {
+        return tf(identity_functor());
     }
 
-    template <class T> 
-    using xtrivially_default_constructible = detail::xtrivial_default_construct_impl<is_complete<std::is_trivially_default_constructible<double>>::value, T>;
+    template <class TF, class FF>
+    auto static_if(std::false_type, const TF&, const FF& ff)
+    {
+        return ff(identity_functor());
+    }
 
-#pragma clang diagnostic pop
-#else
-// CLANG && APPLE
+    template <bool cond, class TF, class FF>
+    auto static_if(const TF& tf, const FF& ff)
+    {
+        return static_if(std::integral_constant<bool, cond>(), tf, ff);
+    }
 
-    template <class T> 
+    /********************************************
+     * xtrivial_default_construct implemenation *
+     ********************************************/
+
+    #if !defined(__GNUG__) || defined(_LIBCPP_VERSION) || defined(_GLIBCXX_USE_CXX11_ABI)
+
+    template <class T>
     using xtrivially_default_constructible = std::is_trivially_default_constructible<T>;
 
-#endif
-#else
-// NOT CLANG
-    #if defined(__GNUC__) && (__GNUC__ < 5 || (__GNUC__ == 5 && __GNUC_MINOR__ < 1))
-    // OLD GCC
-
-    template <class T> 
-    using xtrivially_default_constructible = std::has_trivial_default_constructor<T>;
-
     #else
 
-    template <class T> 
-    using xtrivially_default_constructible = std::is_trivially_default_constructible<T>;
+    template <class T>
+    using xtrivially_default_constructible = std::has_trivial_default_constructor<T>;
 
     #endif
 
-#endif
-    
+    /*************************
+     * conditional type cast *
+     *************************/
+
+    template <bool condition, class T>
+    struct conditional_cast_functor;
+
+    template <class T>
+    struct conditional_cast_functor<false, T>
+    : public identity_functor
+    {
+    };
+
+    template <class T>
+    struct conditional_cast_functor<true, T>
+    {
+        template <class U>
+        inline auto operator()(U && u) const
+        {
+            return static_cast<T>(std::forward<U>(u));
+        }
+    };
+
+    /**
+     * @brief Perform a type cast when a condition is true.
+     * If <tt>condition</tt> is true, return <tt>static_cast<T>(u)</tt>,
+     * otherwise return <tt>u</tt> unchanged. This is useful when an unconditional
+     * static_cast would force undesired type conversions in some situations where
+     * an error or warning would be desired. The condition determines when the
+     * explicit cast is ok.
+     */
+    template <bool condition, class T, class U>
+    inline auto conditional_cast(U && u)
+    {
+        return conditional_cast_functor<condition, T>()(std::forward<U>(u));
+    };
+
+    /************************************
+     * arithmetic type promotion traits *
+     ************************************/
+
+    /**
+     * @brief Traits class for the result type of mixed arithmetic expressions.
+     * For example, <tt>promote_type<unsigned char, unsigned char>::type</tt> tells
+     * the user that <tt>unsigned char + unsigned char => int</tt>.
+     */
+    template <class... T>
+    struct promote_type;
+
+    template <class T>
+    struct promote_type<T>
+    {
+        using type = typename promote_type<T, T>::type;
+    };
+
+    template <class T0, class T1>
+    struct promote_type<T0, T1>
+    {
+        using type = decltype(*(std::decay_t<T0>*)0 + *(std::decay_t<T1>*)0);
+    };
+
+    template <class T0, class... REST>
+    struct promote_type<T0, REST...>
+    {
+        using type = decltype(*(std::decay_t<T0>*)0 + *(typename promote_type<REST...>::type*)0);
+    };
+
+    template <>
+    struct promote_type<bool>
+    {
+        using type = bool;
+    };
+
+    template <class T>
+    struct promote_type<bool, T>
+    {
+        using type = T;
+    };
+
+    template <class... REST>
+    struct promote_type<bool, REST...>
+    {
+        using type = typename promote_type<bool, typename promote_type<REST...>::type>::type;
+    };
+
+    /** 
+     * @brief Abbreviation of 'typename promote_type<T>::type'.
+     */
+    template <class... T>
+    using promote_type_t = typename promote_type<T...>::type;
+
+    /**
+     * @brief Traits class to find the biggest type of the same kind.
+     *
+     * For example, <tt>big_promote_type<unsigned char>::type</tt> is <tt>unsigned long long</tt>.
+     * The default implementation only supports built-in types and <tt>std::complex</tt>. All
+     * other types remain unchanged unless <tt>big_promote_type</tt> gets specialized for them.
+     */
+    template <class T>
+    struct big_promote_type
+    {
+    private:
+        
+        using V = std::decay_t<T>;
+        static constexpr bool is_arithmetic = std::is_arithmetic<V>::value;
+        static constexpr bool is_signed = std::is_signed<V>::value;
+        static constexpr bool is_integral = std::is_integral<V>::value;
+        static constexpr bool is_long_double = std::is_same<V, long double>::value;
+
+    public:
+        using type = std::conditional_t<is_arithmetic,
+                        std::conditional_t<is_integral,
+                            std::conditional_t<is_signed, long long, unsigned long long>,
+                            std::conditional_t<is_long_double, long double, double>
+                        >,
+                        V
+                     >;
+    };
+
+    template <class T>
+    struct big_promote_type<std::complex<T>>
+    {
+        using type = std::complex<typename big_promote_type<T>::type>;
+    };
+
+    /**
+     * @brief Abbreviation of 'typename big_promote_type<T>::type'.
+     */
+    template <class T>
+    using big_promote_type_t = typename big_promote_type<T>::type;
+
+    namespace traits_detail
+    {
+        using std::sqrt;
+
+        template <class T>
+        using real_promote_type_t = decltype(sqrt(*(std::decay_t<T>*)0));
+    }
+
+    /**
+     * @brief Result type of algebraic expressions.
+     *
+     * For example, <tt>real_promote_type<int>::type</tt> tells the
+     * user that <tt>sqrt(int) => double</tt>.
+     */
+    template <class T>
+    struct real_promote_type
+    {
+        using type = traits_detail::real_promote_type_t<T>;
+    };
+
+    /**
+     * @brief Abbreviation of 'typename real_promote_type<T>::type'.
+     */
+    template <class T>
+    using real_promote_type_t = typename real_promote_type<T>::type;
+
+    /**
+     * @brief Traits class to replace 'bool' with 'uint8_t' and keep everything else.
+     *
+     * This is useful for scientific computing, where a boolean mask array is
+     * usually implemented as an array of bytes.
+     */
+    template <class T>
+    struct bool_promote_type
+    {
+        using type = typename std::conditional<std::is_same<T, bool>::value, uint8_t, T>::type;
+    };
+
+    /**
+     * @brief Abbreviation for typename bool_promote_type<T>::type
+     */
+    template <class T>
+    using bool_promote_type_t = typename bool_promote_type<T>::type;
+
+    /********************************************
+     * type inference for norm and squared norm *
+     ********************************************/
+
+    template <class T>
+    struct norm_type;
+
+    template <class T>
+    struct squared_norm_type;
+
+    namespace traits_detail
+    {
+
+        template <class T, bool scalar = std::is_arithmetic<T>::value>
+        struct norm_of_scalar_impl;
+
+        template <class T>
+        struct norm_of_scalar_impl<T, false>
+        {
+            static const bool value = false;
+            using norm_type = void*;
+            using squared_norm_type = void*;
+        };
+
+        template <class T>
+        struct norm_of_scalar_impl<T, true>
+        {
+            static const bool value = true;
+            using norm_type = promote_type_t<T>;
+            using squared_norm_type = promote_type_t<T>;
+        };
+
+        template <class T, bool integral = std::is_integral<T>::value,
+                  bool floating = std::is_floating_point<T>::value>
+        struct norm_of_array_elements_impl;
+
+        template <>
+        struct norm_of_array_elements_impl<void*, false, false>
+        {
+            using norm_type = void*;
+            using squared_norm_type = void*;
+        };
+
+        template <class T>
+        struct norm_of_array_elements_impl<T, false, false>
+        {
+            using norm_type = typename norm_type<T>::type;
+            using squared_norm_type = typename squared_norm_type<T>::type;
+        };
+
+        template <class T>
+        struct norm_of_array_elements_impl<T, true, false>
+        {
+            static_assert(!std::is_same<T, char>::value,
+                          "'char' is not a numeric type, use 'signed char' or 'unsigned char'.");
+
+            using norm_type = double;
+            using squared_norm_type = uint64_t;
+        };
+
+        template <class T>
+        struct norm_of_array_elements_impl<T, false, true>
+        {
+            using norm_type = double;
+            using squared_norm_type = double;
+        };
+
+        template <>
+        struct norm_of_array_elements_impl<long double, false, true>
+        {
+            using norm_type = long double;
+            using squared_norm_type = long double;
+        };
+
+        template <class ARRAY>
+        struct norm_of_vector_impl
+        {
+            static void* test(...);
+
+            template <class U>
+            static typename U::value_type test(U*, typename U::value_type* = 0);
+
+            using T = decltype(test((ARRAY*)0));
+
+            static const bool value = !std::is_same<T, void*>::value;
+
+            using norm_type = typename norm_of_array_elements_impl<T>::norm_type;
+            using squared_norm_type = typename norm_of_array_elements_impl<T>::squared_norm_type;
+        };
+
+        template <class U>
+        struct norm_type_base
+        {
+            using T = std::decay_t<U>;
+
+            static_assert(!std::is_same<T, char>::value,
+                          "'char' is not a numeric type, use 'signed char' or 'unsigned char'.");
+
+            using norm_of_scalar = norm_of_scalar_impl<T>;
+            using norm_of_vector = norm_of_vector_impl<T>;
+
+            static const bool value = norm_of_scalar::value || norm_of_vector::value;
+
+            static_assert(value, "norm_type<T> are undefined for type U.");
+        };
+    }  // namespace traits_detail
+
+    /**
+     * @brief Traits class for the result type of the <tt>norm_l2()</tt> function.
+     *
+     * Member 'type' defines the result of <tt>norm_l2(t)</tt>, where <tt>t</tt>
+     * is of type @tparam T. It implements the following rules designed to
+     * minimize the potential for overflow:
+     *   - @tparam T is an arithmetic type: 'type' is the result type of <tt>abs(t)</tt>.
+     *   - @tparam T is a container of 'long double' elements: 'type' is <tt>long double</tt>.
+     *   - @tparam T is a container of another arithmetic type: 'type' is <tt>double</tt>.
+     *   - @tparam T is a container of some other type: 'type' is the element's norm type,
+     *
+     * Containers are recognized by having an embedded typedef 'value_type'.
+     * To change the behavior for a case not covered here, specialize the
+     * <tt>traits_detail::norm_type_base</tt> template.
+     */
+    template <class T>
+    struct norm_type
+        : public traits_detail::norm_type_base<T>
+    {
+        using base_type = traits_detail::norm_type_base<T>;
+
+        using type =
+            typename std::conditional<base_type::norm_of_vector::value,
+                                      typename base_type::norm_of_vector::norm_type,
+                                      typename base_type::norm_of_scalar::norm_type>::type;
+    };
+
+    /**
+     * Abbreviation of 'typename norm_type<T>::type'.
+     */
+    template <class T>
+    using norm_type_t = typename norm_type<T>::type;
+
+    /**
+     * @brief Traits class for the result type of the <tt>norm_sq()</tt> function.
+     *
+     * Member 'type' defines the result of <tt>norm_sq(t)</tt>, where <tt>t</tt>
+     * is of type @tparam T. It implements the following rules designed to
+     * minimize the potential for overflow:
+     *   - @tparam T is an arithmetic type: 'type' is the result type of <tt>t*t</tt>.
+     *   - @tparam T is a container of 'long double' elements: 'type' is <tt>long double</tt>.
+     *   - @tparam T is a container of another floating-point type: 'type' is <tt>double</tt>.
+     *   - @tparam T is a container of integer elements: 'type' is <tt>uint64_t</tt>.
+     *   - @tparam T is a container of some other type: 'type' is the element's squared norm type,
+     *
+     *  Containers are recognized by having an embedded typedef 'value_type'.
+     *  To change the behavior for a case not covered here, specialize the
+     *  <tt>traits_detail::norm_type_base</tt> template.
+     */
+    template <class T>
+    struct squared_norm_type
+        : public traits_detail::norm_type_base<T>
+    {
+        using base_type = traits_detail::norm_type_base<T>;
+
+        using type =
+            typename std::conditional<base_type::norm_of_vector::value,
+                                      typename base_type::norm_of_vector::squared_norm_type,
+                                      typename base_type::norm_of_scalar::squared_norm_type>::type;
+    };
+
+    /**
+     * Abbreviation of 'typename squared_norm_type<T>::type'.
+     */
+    template <class T>
+    using squared_norm_type_t = typename squared_norm_type<T>::type;
 }
 
 #endif
