@@ -6,8 +6,8 @@
 * The full license is in the file LICENSE, distributed with this software. *
 ****************************************************************************/
 
-#ifndef XVIEW_HPP
-#define XVIEW_HPP
+#ifndef XTENSOR_VIEW_HPP
+#define XTENSOR_VIEW_HPP
 
 #include <algorithm>
 #include <array>
@@ -16,6 +16,9 @@
 #include <type_traits>
 #include <utility>
 
+#include "xtl/xclosure.hpp"
+#include "xtl/xsequence.hpp"
+
 #include "xbroadcast.hpp"
 #include "xcontainer.hpp"
 #include "xiterable.hpp"
@@ -48,12 +51,8 @@ namespace xt
     {
         using xexpression_type = std::decay_t<CT>;
         using inner_shape_type = typename xview_shape_type<typename xexpression_type::shape_type, S...>::type;
-        using stepper = xview_stepper<false, CT, S...>;
-        using const_stepper = xview_stepper<true, CT, S...>;
-        using iterator = xiterator<stepper, inner_shape_type*, DEFAULT_LAYOUT>;
-        using const_iterator = xiterator<const_stepper, inner_shape_type*, DEFAULT_LAYOUT>;
-        using reverse_iterator = std::reverse_iterator<iterator>;
-        using const_reverse_iterator = std::reverse_iterator<const_iterator>;
+        using stepper = xview_stepper<std::is_const<std::remove_reference_t<CT>>::value, CT, S...>;
+        using const_stepper = xview_stepper<true, std::remove_cv_t<CT>, S...>;
     };
 
     /**
@@ -72,9 +71,8 @@ namespace xt
      */
     template <class CT, class... S>
     class xview : public xview_semantic<xview<CT, S...>>,
-                  public xexpression_iterable<xview<CT, S...>>
+                  public xiterable<xview<CT, S...>>
     {
-
     public:
 
         using self_type = xview<CT, S...>;
@@ -89,7 +87,7 @@ namespace xt
         using size_type = typename xexpression_type::size_type;
         using difference_type = typename xexpression_type::difference_type;
 
-        using iterable_base = xexpression_iterable<self_type>;
+        using iterable_base = xiterable<self_type>;
         using inner_shape_type = typename iterable_base::inner_shape_type;
         using shape_type = inner_shape_type;
         using strides_type = shape_type;
@@ -102,8 +100,13 @@ namespace xt
         static constexpr layout_type static_layout = layout_type::dynamic;
         static constexpr bool contiguous_layout = false;
 
-        template <class CTA, class... SL>
-        explicit xview(CTA&& e, SL&&... slices) noexcept;
+        // The FSL argument prevents the compiler from calling this constructor
+        // instead of the copy constructor when sizeof...(SL) == 0.
+        template <class CTA, class FSL, class... SL>
+        explicit xview(CTA&& e, FSL&& first_slice, SL&&... slices) noexcept;
+
+        xview(const xview&) = default;
+        self_type& operator=(const xview& rhs);
 
         template <class E>
         self_type& operator=(const xexpression<E>& e);
@@ -120,6 +123,8 @@ namespace xt
 
         template <class... Args>
         reference operator()(Args... args);
+        template <class... Args>
+        reference at(Args... args);
         reference operator[](const xindex& index);
         reference operator[](size_type i);
         template <class It>
@@ -127,6 +132,8 @@ namespace xt
 
         template <class... Args>
         const_reference operator()(Args... args) const;
+        template <class... Args>
+        const_reference at(Args... args) const;
         const_reference operator[](const xindex& index) const;
         const_reference operator[](size_type i) const;
         template <class It>
@@ -141,12 +148,12 @@ namespace xt
         template <class ST>
         stepper stepper_begin(const ST& shape);
         template <class ST>
-        stepper stepper_end(const ST& shape);
+        stepper stepper_end(const ST& shape, layout_type l);
 
         template <class ST>
         const_stepper stepper_begin(const ST& shape) const;
         template <class ST>
-        const_stepper stepper_end(const ST& shape) const;
+        const_stepper stepper_end(const ST& shape, layout_type l) const;
 
         template <class T = xexpression_type>
         std::enable_if_t<has_raw_data_interface<T>::value, const typename T::container_type&>
@@ -170,6 +177,10 @@ namespace xt
 
         size_type underlying_size(size_type dim) const;
 
+        xtl::xclosure_pointer<self_type&> operator&() &;
+        xtl::xclosure_pointer<const self_type&> operator&() const&;
+        xtl::xclosure_pointer<self_type> operator&() && ;
+
     private:
 
         // VS 2015 workaround (yes, really)
@@ -210,7 +221,7 @@ namespace xt
         template <class It>
         base_index_type make_index(It first, It last) const;
 
-        void assign_temporary_impl(temporary_type& tmp);
+        void assign_temporary_impl(temporary_type&& tmp);
 
         friend class xview_semantic<xview<CT, S...>>;
     };
@@ -245,7 +256,6 @@ namespace xt
     template <bool is_const, class CT, class... S>
     class xview_stepper
     {
-
     public:
 
         using view_type = std::conditional_t<is_const,
@@ -273,7 +283,7 @@ namespace xt
         void reset_back(size_type dim);
 
         void to_begin();
-        void to_end();
+        void to_end(layout_type);
 
         bool equal(const xview_stepper& rhs) const;
 
@@ -328,14 +338,15 @@ namespace xt
      * Users should not call directly this constructor but
      * use the view function instead.
      * @param e the xexpression to adapt
+     * @param first_slice the first slice describing the view
      * @param slices the slices list describing the view
      * @sa view
      */
     template <class CT, class... S>
-    template <class CTA, class... SL>
-    inline xview<CT, S...>::xview(CTA&& e, SL&&... slices) noexcept
-        : m_e(std::forward<CTA>(e)), m_slices(std::forward<SL>(slices)...),
-          m_shape(make_sequence<shape_type>(m_e.dimension() - integral_count<S...>() + newaxis_count<S...>(), 0))
+    template <class CTA, class FSL, class... SL>
+    inline xview<CT, S...>::xview(CTA&& e, FSL&& first_slice, SL&&... slices) noexcept
+        : m_e(std::forward<CTA>(e)), m_slices(std::forward<FSL>(first_slice), std::forward<SL>(slices)...),
+          m_shape(xtl::make_sequence<shape_type>(m_e.dimension() - integral_count<S...>() + newaxis_count<S...>(), 0))
     {
         auto func = [](const auto& s) noexcept { return get_size(s); };
         for (size_type i = 0; i != dimension(); ++i)
@@ -347,6 +358,13 @@ namespace xt
     }
     //@}
 
+    template <class CT, class... S>
+    inline auto xview<CT, S...>::operator=(const xview& rhs) -> self_type&
+    {
+        temporary_type tmp(rhs);
+        return this->assign_temporary(std::move(tmp));
+    }
+
     /**
      * @name Extended copy semantic
      */
@@ -419,7 +437,7 @@ namespace xt
     {
         return m_slices;
     }
-        
+
     /**
      * Returns the slices of the view.
      */
@@ -444,10 +462,29 @@ namespace xt
     template <class... Args>
     inline auto xview<CT, S...>::operator()(Args... args) -> reference
     {
+        // The static cast prevents the compiler from instantiating the template methods with signed integers,
+        // leading to warning about signed/unsigned conversions in the deeper layers of the access methods
         return access_impl(std::make_index_sequence<(sizeof...(Args) + integral_count<S...>() > newaxis_count<S...>() ?
                                                         sizeof...(Args) + integral_count<S...>() - newaxis_count<S...>() :
                                                         0)>(),
-                           args...);
+                           static_cast<size_type>(args)...);
+    }
+
+    /**
+     * Returns a reference to the element at the specified position in the expression,
+     * after dimension and bounds checking.
+     * @param args a list of indices specifying the position in the function. Indices
+     * must be unsigned integers, the number of indices should be equal to the number of dimensions
+     * of the expression.
+     * @exception std::out_of_range if the number of argument is greater than the number of dimensions
+     * or if indices are out of bounds.
+     */
+    template <class CT, class... S>
+    template <class... Args>
+    inline auto xview<CT, S...>::at(Args... args) -> reference
+    {
+        check_access(shape(), static_cast<size_type>(args)...);
+        return this->operator()(args...);
     }
 
     template <class CT, class... S>
@@ -466,6 +503,7 @@ namespace xt
     template <class It>
     inline auto xview<CT, S...>::element(It first, It last) -> reference
     {
+        // TODO: avoid memory allocation
         auto index = make_index(first, last);
         return m_e.element(index.cbegin(), index.cend());
     }
@@ -480,10 +518,29 @@ namespace xt
     template <class... Args>
     inline auto xview<CT, S...>::operator()(Args... args) const -> const_reference
     {
+        // The static cast prevents the compiler from instantiating the template methods with signed integers,
+        // leading to warning about signed/unsigned conversions in the deeper layers of the access methods
         return access_impl(std::make_index_sequence<(sizeof...(Args) + integral_count<S...>() > newaxis_count<S...>() ?
                                                         sizeof...(Args) + integral_count<S...>() - newaxis_count<S...>() :
                                                         0)>(),
-                           args...);
+                           static_cast<size_type>(args)...);
+    }
+
+    /**
+     * Returns a constant reference to the element at the specified position in the expression,
+     * after dimension and bounds checking.
+     * @param args a list of indices specifying the position in the function. Indices
+     * must be unsigned integers, the number of indices should be equal to the number of dimensions
+     * of the expression.
+     * @exception std::out_of_range if the number of argument is greater than the number of dimensions
+     * or if indices are out of bounds.
+     */
+    template <class CT, class... S>
+    template <class... Args>
+    inline auto xview<CT, S...>::at(Args... args) const -> const_reference
+    {
+        check_access(shape(), static_cast<size_type>(args)...);
+        return this->operator()(args...);
     }
 
     template <class CT, class... S>
@@ -502,6 +559,7 @@ namespace xt
     template <class It>
     inline auto xview<CT, S...>::element(It first, It last) const -> const_reference
     {
+        // TODO: avoid memory allocation
         auto index = make_index(first, last);
         return m_e.element(index.cbegin(), index.cend());
     }
@@ -528,13 +586,13 @@ namespace xt
         std::enable_if_t<has_raw_data_interface<T>::value, const typename T::strides_type>
     {
         using strides_type = typename T::strides_type;
-        strides_type temp = m_e.strides();
-        strides_type strides = make_sequence<strides_type>(m_e.dimension() - integral_count<S...>(), 0);
+        strides_type strides = xtl::make_sequence<strides_type>(m_e.dimension() - integral_count<S...>(), 0);
 
         auto func = [](const auto& s) { return xt::step_size(s); };
         size_type i = 0, idx;
 
-        for (; i < sizeof...(S); ++i) {
+        for (; i < sizeof...(S); ++i)
+        {
             idx = integral_skip<S...>(i);
             if (idx >= sizeof...(S))
             {
@@ -542,7 +600,8 @@ namespace xt
             }
             strides[i] = m_e.strides()[idx] * apply<size_type>(idx, func, m_slices);
         }
-        for (; i < strides.size(); ++i) {
+        for (; i < strides.size(); ++i)
+        {
             strides[i] = m_e.strides()[idx++];
         }
 
@@ -594,6 +653,24 @@ namespace xt
         return m_e.shape()[dim];
     }
 
+    template <class CT, class... S>
+    inline auto xview<CT, S...>::operator&() & -> xtl::xclosure_pointer<self_type&>
+    {
+        return xtl::closure_pointer(*this);
+    }
+
+    template <class CT, class... S>
+    inline auto xview<CT, S...>::operator&() const& -> xtl::xclosure_pointer<const self_type&>
+    {
+        return xtl::closure_pointer(*this);
+    }
+
+    template <class CT, class... S>
+    inline auto xview<CT, S...>::operator&() && ->xtl::xclosure_pointer<self_type>
+    {
+        return xtl::closure_pointer(std::move(*this));
+    }
+
     /**
      * @name Broadcasting
      */
@@ -663,21 +740,22 @@ namespace xt
     template <typename std::decay_t<CT>::size_type I, class T, class Arg, class... Args>
     inline auto xview<CT, S...>::sliced_access(const xslice<T>& slice, Arg arg, Args... args) const -> size_type
     {
-        return slice.derived_cast()(argument<I>(arg, args...));
+        using ST = typename T::size_type;
+        return static_cast<size_type>(slice.derived_cast()(argument<I>(static_cast<ST>(arg), static_cast<ST>(args)...)));
     }
 
     template <class CT, class... S>
     template <typename std::decay_t<CT>::size_type I, class T, class... Args>
     inline auto xview<CT, S...>::sliced_access(const T& squeeze, Args...) const -> disable_xslice<T, size_type>
     {
-        return squeeze;
+        return static_cast<size_type>(squeeze);
     }
 
     template <class CT, class... S>
     template <class It>
     inline auto xview<CT, S...>::make_index(It first, It last) const -> base_index_type
     {
-        auto index = make_sequence<typename xexpression_type::shape_type>(m_e.dimension(), 0);
+        auto index = xtl::make_sequence<typename xexpression_type::shape_type>(m_e.dimension(), 0);
         auto func1 = [&first](const auto& s)
         {
             return get_slice_value(s, first);
@@ -705,9 +783,9 @@ namespace xt
     }
 
     template <class CT, class... S>
-    inline void xview<CT, S...>::assign_temporary_impl(temporary_type& tmp)
+    inline void xview<CT, S...>::assign_temporary_impl(temporary_type&& tmp)
     {
-        std::copy(tmp.cbegin(), tmp.cend(), this->xbegin());
+        std::copy(tmp.cbegin(), tmp.cend(), this->begin());
     }
 
     namespace detail
@@ -721,7 +799,7 @@ namespace xt
         template <class E, std::size_t... I, class... S>
         inline auto make_view_impl(E&& e, std::index_sequence<I...>, S&&... slices)
         {
-            using view_type = xview<closure_t<E>, get_slice_type<std::decay_t<E>, S>...>;
+            using view_type = xview<xtl::closure_type_t<E>, get_slice_type<std::decay_t<E>, S>...>;
             return view_type(std::forward<E>(e),
                 get_slice_implementation(e, std::forward<S>(slices), get_underlying_shape_index<std::decay_t<E>, S...>(I))...
             );
@@ -756,10 +834,10 @@ namespace xt
 
     template <class CT, class... S>
     template <class ST>
-    inline auto xview<CT, S...>::stepper_end(const ST& shape) -> stepper
+    inline auto xview<CT, S...>::stepper_end(const ST& shape, layout_type l) -> stepper
     {
         size_type offset = shape.size() - dimension();
-        return stepper(this, m_e.stepper_end(m_e.shape()), offset, true);
+        return stepper(this, m_e.stepper_end(m_e.shape(), l), offset, true);
     }
 
     template <class CT, class... S>
@@ -773,11 +851,11 @@ namespace xt
 
     template <class CT, class... S>
     template <class ST>
-    inline auto xview<CT, S...>::stepper_end(const ST& shape) const -> const_stepper
+    inline auto xview<CT, S...>::stepper_end(const ST& shape, layout_type l) const -> const_stepper
     {
         size_type offset = shape.size() - dimension();
         const xexpression_type& e = m_e;
-        return const_stepper(this, e.stepper_end(m_e.shape()), offset, true);
+        return const_stepper(this, e.stepper_end(m_e.shape(), l), offset, true);
     }
 
     /********************************
@@ -849,9 +927,9 @@ namespace xt
     }
 
     template <bool is_const, class CT, class... S>
-    inline void xview_stepper<is_const, CT, S...>::to_end()
+    inline void xview_stepper<is_const, CT, S...>::to_end(layout_type l)
     {
-        m_it.to_end();
+        m_it.to_end(l);
         to_end_impl();
     }
 
